#-*- coding:utf-8 -*-
#
# simulation/nest_plot.py
#
# This file is part of the NNGT project, a graph-library for standardized and
# and reproducible graph analysis: generate and analyze networks with your
# favorite graph library (graph-tool/igraph/networkx) on any platform, without
# any change to your code.
# Copyright (C) 2015-2021 Tanguy Fardet
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

# nest_plot.py

# This file is part of the NNGT module
# Distributed as a free software, in the hope that it will be useful, under the
# terms of the GNU General Public License.

""" Utility functions to plot NEST simulated activity """

import itertools
import logging

from matplotlib.colors import ColorConverter
import numpy as np
import nest

import nngt

from nngt.analysis import total_firing_rate
from nngt.lib import InvalidArgument, nonstring_container, is_integer
from nngt.lib.sorting import _sort_groups, _sort_neurons
from nngt.lib.logger import _log_message
from nngt.plot import palette_discrete, markers
from nngt.plot.plt_properties import _set_new_plot, _set_ax_lims

from .nest_utils import nest_version, spike_rec, _get_nest_gids

logger = logging.getLogger(__name__)


# --------------------- #
# Plotting the activity #
# --------------------- #

def plot_activity(gid_recorder=None, record=None, network=None, gids=None,
                  axis=None, show=False, limits=None, histogram=False,
                  title=None, fignum=None, label=None, sort=None,
                  average=False, normalize=1., decimate=None, transparent=True,
                  kernel_center=0., kernel_std=None, resolution=None,
                  cut_gaussian=5., **kwargs):
    '''
    Plot the monitored activity.

    .. versionchanged:: 1.2
        Switched `hist` to `histogram` and default value to False.

    .. versionchanged:: 1.0.1
        Added `axis` parameter, restored missing `fignum` parameter.

    Parameters
    ----------
    gid_recorder : tuple or list of tuples, optional (default: None)
        The gids of the recording devices. If None, then all existing
        spike_recs are used.
    record : tuple or list, optional (default: None)
        List of the monitored variables for each device. If `gid_recorder` is
        None, record can also be None and only spikes are considered.
    network : :class:`~nngt.Network` or subclass, optional (default: None)
        Network which activity will be monitored.
    gids : tuple, optional (default: None)
        NEST gids of the neurons which should be monitored.
    axis : matplotlib axis object, optional (default: new one)
        Axis that should be use to plot the activity. This takes precedence
        over `fignum`.
    show : bool, optional (default: False)
        Whether to show the plot right away or to wait for the next plt.show().
    histogram : bool, optional (default: False)
        Whether to display the histogram when plotting spikes rasters.
    limits : tuple, optional (default: None)
        Time limits of the plot (if not specified, times of first and last
        spike for raster plots).
    title : str, optional (default: None)
        Title of the plot.
    fignum : int, or dict, optional (default: None)
        Plot the activity on an existing figure (from ``figure.number``). This
        parameter is ignored if `axis` is provided.
    label : str or list, optional (default: None)
        Add labels to the plot (one per recorder).
    sort : str or list, optional (default: None)
        Sort neurons using a topological property ("in-degree", "out-degree",
        "total-degree" or "betweenness"), an activity-related property
        ("firing_rate" or neuronal property) or a user-defined list of sorted
        neuron ids. Sorting is performed by increasing value of the `sort`
        property from bottom to top inside each group.
    normalize : float or list, optional (default: None)
        Normalize the recorded results by a given float. If a list is provided,
        there should be one entry per voltmeter or multimeter in the recorders.
        If the recording was done through `monitor_groups`, the population can
        be passed to normalize the data by the nuber of nodes in each group.
    decimate : int or list of ints, optional (default: None)
        Represent only a fraction of the spiking neurons; only one neuron in
        `decimate` will be represented (e.g. setting `decimate` to 5 will lead
        to only 20% of the neurons being represented). If a list is provided,
        it must have one entry per NeuralGroup in the population.
    kernel_center : float, optional (default: 0.)
        Temporal shift of the Gaussian kernel, in ms (for the histogram).
    kernel_std : float, optional (default: 0.5% of simulation time)
        Characteristic width of the Gaussian kernel (standard deviation) in ms
        (for the histogram).
    resolution : float or array, optional (default: `0.1*kernel_std`)
        The resolution at which the firing rate values will be computed.
        Choosing a value smaller than `kernel_std` is strongly advised.
        If resolution is an array, it will be considered as the times were the
        firing rate should be computed (for the histogram).
    cut_gaussian : float, optional (default: 5.)
        Range over which the Gaussian will be computed (for the histogram).
        By default, we consider the 5-sigma range. Decreasing this value will
        increase speed at the cost of lower fidelity; increasing it with
        increase the fidelity at the cost of speed.
    **kwargs : dict
        "color" and "alpha" values can be overriden here.

    Warning
    -------
    Sorting with "firing_rate" only works if NEST gids form a continuous
    integer range.

    Returns
    -------
    lines : list of lists of :class:`matplotlib.lines.Line2D`
        Lines containing the data that was plotted, grouped by figure.
    '''
    import matplotlib.pyplot as plt
    recorders = _get_nest_gids([])
    lst_labels, lines, axes, labels = [], {}, {}, {}

    # normalize recorders and recordables
    if gid_recorder is not None:
        assert record is not None, "`record` must also be provided."
        if len(record) != len(gid_recorder):
            raise InvalidArgument('`record` must either be the same for all '
                                  'recorders, or contain one entry per '
                                  'recorder in `gid_recorder`')
        for rec in gid_recorder:
            if nest_version == 3:
                recorders = _get_nest_gids(gid_recorder)
            else:
                if isinstance(gid_recorder[0], tuple):
                    recorders.append(rec)
                else:
                    recorders.append((rec,))
    else:
        prop = {'model': spike_rec}
        if nest_version == 3:
            recorders = nest.GetNodes(properties=prop)
        else:
            recorders = [
                (gid,) for gid in nest.GetNodes((0,), properties=prop)[0]
            ]

        record = tuple("spikes" for _ in range(len(recorders)))

    # get gids and groups
    gids = network.nest_gids if (gids is None and network is not None) \
           else gids

    if gids is None:
        gids = []

        for rec in recorders:
            gids.extend(nest.GetStatus(rec)[0]["events"]["senders"])

        gids = np.unique(gids)

    num_group = 1 if network is None else len(network.population)
    num_lines = max(num_group, len(recorders))

    # sorting
    sorted_neurons = np.array([])

    if len(gids):
        sorted_neurons = np.arange(
            np.max(gids) + 1).astype(int) - np.min(gids) + 1

    attr = None

    if sort is not None:
        assert network is not None, "`network` is required for sorting."
        if nonstring_container(sort):
            attr = sort
            sorted_neurons = _sort_neurons(attr, gids, network)
            sort = "user defined sort"
        else:
            data = None
            if sort.lower() in ("firing_rate", "b2"):  # get senders
                data = [[], []]
                for rec in recorders:
                    info = nest.GetStatus(rec)[0]
                    if str(info["model"]) == spike_rec:
                        data[0].extend(info["events"]["senders"])
                        data[1].extend(info["events"]["times"])
                data = np.array(data).T
            sorted_neurons, attr = _sort_neurons(
                sort, gids, network, data=data, return_attr=True)
    elif network is not None and network.is_spatial():
        sorted_neurons, attr = _sort_neurons(
            "space", gids, network, data=None, return_attr=True)

    # spikes plotting
    colors = palette_discrete(np.linspace(0, 1, num_lines))
    num_raster, num_detec, num_meter = 0, 0, 0
    fignums = fignum if isinstance(fignum, dict) else {}
    decim = []
    if decimate is None:
        decim = [None for _ in range(num_lines)]
    elif is_integer(decimate):
        decim = [decimate for _ in range(num_lines)]
    elif nonstring_container(decimate):
        assert len(decimate) == num_lines, "`decimate` should have one " +\
                                           "entry per plot."
        decim = decimate
    else:
        raise AttributeError(
            "`decimate` must be either an int or a list of `int`.")

    # set labels
    if label is None:
        lst_labels = [None for _ in range(len(recorders))]
    else:
        if isinstance(label, str):
            lst_labels = [label]
        else:
            lst_labels = label
        if len(label) != len(recorders):
            _log_message(logger, "WARNING",
                         'Incorrect length for `label`: expecting {} but got '
                         '{}.\nIgnoring.'.format(len(recorders), len(label)))
            lst_labels = [None for _ in range(len(recorders))]

    datasets = []
    max_time = 0.

    for rec in recorders:
        info = nest.GetStatus(rec)[0]

        if len(info["events"]["times"]):
            max_time = max(max_time, np.max(info["events"]["times"]))

        datasets.append(info)

    if kernel_std is None:
        kernel_std = max_time*0.005

    if resolution is None:
        resolution = 0.5*kernel_std

    # plot
    for info, var, lbl in zip(datasets, record, lst_labels):
        fnum = fignums.get(info["model"], fignum)
        if info["model"] not in labels:
            labels[info["model"]] = []
            lines[info["model"]] = []

        if str(info["model"]) == spike_rec:
            if spike_rec in axes:
                axis = axes[spike_rec]
            c = colors[num_raster]
            times, senders = info["events"]["times"], info["events"]["senders"]
            sorted_ids = sorted_neurons[senders]
            l = raster_plot(times, sorted_ids, color=c, show=False,
                            limits=limits, sort=sort, fignum=fnum, axis=axis,
                            decimate=decim[num_raster], sort_attribute=attr,
                            network=network, histogram=histogram,
                            transparent=transparent,
                            hist_ax=axes.get('histogram', None),
                            kernel_center=kernel_center,
                            kernel_std=kernel_std, resolution=resolution,
                            cut_gaussian=cut_gaussian)
            num_raster += 1
            if l:
                fig_raster = l[0].figure.number
                fignums[spike_rec] = fig_raster
                axes[spike_rec] = l[0].axes
                labels[spike_rec].append(lbl)
                lines[spike_rec].extend(l)
                if histogram:
                    axes['histogram'] = l[1].axes
        elif "detector" in str(info["model"]):
            c = colors[num_detec]
            times, senders = info["events"]["times"], info["events"]["senders"]
            sorted_ids = sorted_neurons[senders]
            l = raster_plot(times, sorted_ids, fignum=fnum, color=c, axis=axis,
                            show=False, histogram=histogram, limits=limits,
                            kernel_center=kernel_center,
                            kernel_std=kernel_std, resolution=resolution,
                            cut_gaussian=cut_gaussian)
            if l:
                fig_detect = l[0].figure.number
                num_detec += 1
                fignums[info["model"]] = fig_detect
                labels[info["model"]].append(lbl)
                lines[info["model"]].extend(l)
                if histogram:
                    axes['histogram'] = l[1].axes
        else:
            da_time  = info["events"]["times"]
            # prepare axis setup
            fig = None
            if axis is None:
                fig = plt.figure(fnum)
                fignums[info["model"]] = fig.number
            else:
                fig = axis.get_figure()
            lines_tmp, labels_tmp = [], []
            if nonstring_container(var):
                m_colors = palette_discrete(np.linspace(0, 1, len(var)))
                axes = fig.axes
                if axis is not None:
                    # multiple y axes on a single subplot, adapted from
                    # https://matplotlib.org/examples/pylab_examples/
                    # multiple_yaxis_with_spines.html
                    axes = [axis]
                    axis.name = var[0]
                    if len(var) > 1:
                        axes.append(axis.twinx())
                        axes[-1].name = var[1]
                    if len(var) > 2:
                        fig.subplots_adjust(right=0.75)
                        for i, name in zip(range(len(var)-2), var[2:]):
                            new_ax = axis.twinx()
                            new_ax.spines["right"].set_position(
                                ("axes", 1.2*(i+1)))
                            axes.append(new_ax)
                            _make_patch_spines_invisible(new_ax)
                            new_ax.spines["right"].set_visible(True)
                            axes[-1].name = name
                if not axes:
                    axes = _set_new_plot(fig.number, names=var)[1]
                labels_tmp = [lbl for _ in range(len(var))]
                for subvar, c in zip(var, m_colors):
                    c = kwargs.get('color', c)
                    alpha = kwargs.get('alpha', 1)
                    for ax in axes:
                        if ax.name == subvar:
                            da_subvar = info["events"][subvar]
                            if isinstance(normalize, nngt.NeuralPop):
                                da_subvar /= normalize[num_meter].size
                            elif nonstring_container(normalize):
                                da_subvar /= normalize[num_meter]
                            elif normalize is not None:
                                da_subvar /= normalize
                            lines_tmp.extend(
                                ax.plot(da_time, da_subvar, color=c,
                                        alpha=alpha))
                            ax.set_ylabel(subvar)
                            ax.set_xlabel("time")
                            if limits is not None:
                                ax.set_xlim(limits[0], limits[1])
            else:
                num_axes, ax = len(fig.axes), axis
                if axis is None:
                    ax = fig.add_subplot(num_axes + 1, 1, num_axes + 1)
                da_var = info["events"][var]
                c = kwargs.get('color', None)
                alpha = kwargs.get('alpha', 1)
                lines_tmp.extend(ax.plot(da_time, da_var/normalize, color=c,
                                         alpha=alpha))
                labels_tmp.append(lbl)
                ax.set_ylabel(var)
                ax.set_xlabel("time")
            labels[info["model"]].extend(labels_tmp)
            lines[info["model"]].extend(lines_tmp)
            num_meter += 1

    if spike_rec in axes:
        ax = axes[spike_rec]

        if limits is not None:
            ax.set_xlim(limits[0], limits[1])
        else:
            t_min, t_max, idx_min, idx_max = np.inf, -np.inf, np.inf, -np.inf

            for l in ax.lines:
                t_max = max(np.max(l.get_xdata()), t_max)
                t_min = min(np.min(l.get_xdata()), t_max)
                idx_min = min(np.min(l.get_ydata()), idx_min)
                idx_max = max(np.max(l.get_ydata()), idx_max)

            dt   = t_max - t_min
            didx = idx_max - idx_min
            pc   = 0.02

            if not np.any(np.isinf((t_max, t_min))):
                ax.set_xlim([t_min - pc*dt, t_max + pc*dt])

            if not np.any(np.isinf((idx_min, idx_max))):
              ax.set_ylim([idx_min - pc*didx, idx_max + pc*didx])

    for recorder in fignums:
        fig = plt.figure(fignums[recorder])
        if title is not None:
            fig.suptitle(title)
        if label is not None:
            fig.legend(lines[recorder], labels[recorder])

    if show:
        plt.show()

    return lines


def raster_plot(times, senders, limits=None, title="Spike raster",
                histogram=False, num_bins=1000, color="b", decimate=None,
                axis=None, fignum=None, label=None, show=True, sort=None,
                sort_attribute=None, network=None, transparent=True,
                kernel_center=0., kernel_std=30., resolution=None,
                cut_gaussian=5., **kwargs):
    """
    Plotting routine that constructs a raster plot along with
    an optional histogram.

    .. versionchanged:: 1.2
        Switched `hist` to `histogram`.

    .. versionchanged:: 1.0.1
        Added `axis` parameter.

    Parameters
    ----------
    times : list or :class:`numpy.ndarray`
        Spike times.
    senders : list or :class:`numpy.ndarray`
        Index for the spiking neuron for each time in `times`.
    limits : tuple, optional (default: None)
        Time limits of the plot (if not specified, times of first and last
        spike).
    title : string, optional (default: 'Spike raster')
        Title of the raster plot.
    histogram : bool, optional (default: True)
        Whether to plot the raster's histogram.
    num_bins : int, optional (default: 1000)
        Number of bins for the histogram.
    color : string or float, optional (default: 'b')
        Color of the plot lines and markers.
    decimate : int, optional (default: None)
        Represent only a fraction of the spiking neurons; only one neuron in
        `decimate` will be represented (e.g. setting `decimate` to 10 will lead
        to only 10% of the neurons being represented).
    axis : matplotlib axis object, optional (default: new one)
        Axis that should be use to plot the activity.
    fignum : int, optional (default: None)
        Id of another raster plot to which the new data should be added.
    label : str, optional (default: None)
        Label the current data.
    show : bool, optional (default: True)
        Whether to show the plot right away or to wait for the next plt.show().
    kernel_center : float, optional (default: 0.)
        Temporal shift of the Gaussian kernel, in ms.
    kernel_std : float, optional (default: 30.)
        Characteristic width of the Gaussian kernel (standard deviation) in ms.
    resolution : float or array, optional (default: `0.1*kernel_std`)
        The resolution at which the firing rate values will be computed.
        Choosing a value smaller than `kernel_std` is strongly advised.
        If resolution is an array, it will be considered as the times were the
        firing rate should be computed.
    cut_gaussian : float, optional (default: 5.)
        Range over which the Gaussian will be computed (for the histogram).
        By default, we consider the 5-sigma range. Decreasing this value will
        increase speed at the cost of lower fidelity; increasing it with
        increase the fidelity at the cost of speed.

    Returns
    -------
    lines : list of :class:`matplotlib.lines.Line2D`
        Lines containing the data that was plotted.
    """
    import matplotlib.pyplot as plt

    lines = []

    mpl_kwargs = {k: v for k, v in kwargs.items() if k != 'hist_ax'}

    if label is None:
        mpl_kwargs['label'] = label

    # decimate if necessary
    if decimate is not None:
        idx_keep = np.where(np.mod(senders, decimate) == 0)[0]
        senders = senders[idx_keep]
        times = times[idx_keep]

    if len(times):
        if axis is not None:
            fig = axis.get_figure()
        else:
            fig = plt.figure(fignum)
        if transparent:
            fig.patch.set_visible(False)
        ylabel = "Neuron ID"
        xlabel = "Time (ms)"

        delta_t = 0.01*(times[-1]-times[0])

        if histogram:
            ax1, ax2 = None, None
            if kwargs.get("hist_ax", None) is None:
                num_axes = len(fig.axes)
                for i, old_ax in enumerate(fig.axes):
                    old_ax.change_geometry(num_axes + 2, 1, i+1)
                ax1 = fig.add_subplot(num_axes + 2, 1, num_axes + 1)
                ax2 = fig.add_subplot(num_axes + 2, 1, num_axes + 2,
                                      sharex=ax1)
            else:
                ax1 = axis
                ax2 = kwargs["hist_ax"]

            if limits is not None:
                start, stop = limits

                keep    = (times >= start)&(times <= stop)
                times   = times[keep]
                senders = senders[keep]

            lines.extend(ax1.plot(
                times, senders, c=color, marker="o", linestyle='None',
                mec="k", mew=0.5, ms=4, **mpl_kwargs))

            ax1_lines = ax1.lines

            if len(ax1_lines) > 1:
                t_max = max(ax1_lines[0].get_xdata().max(),times[-1])
                ax1.set_xlim([-delta_t, t_max+delta_t])

            ax1.set_ylabel(ylabel)

            if limits is not None:
                ax1.set_xlim(*limits)

            fr, fr_times = total_firing_rate(
                data=np.array([senders, times]).T, kernel_center=kernel_center,
                kernel_std=kernel_std, resolution=resolution,
                cut_gaussian=cut_gaussian)

            hist_lines = ax2.get_lines()

            if hist_lines:
                data = hist_lines[-1].get_data()
                bottom = data[1]
                if limits is None:
                    dt = fr_times[1] - fr_times[0]
                    old_times = data[0]
                    old_start = int(old_times[0] / dt)
                    new_start = int(fr_times[0] / dt)
                    old_end = int(old_times[-1] / dt)
                    new_end = int(fr_times[-1] / dt)
                    diff_start = new_start-old_start
                    diff_end = new_end-old_end
                    if diff_start > 0:
                        bottom = bottom[diff_start:]
                    else:
                        bottom = np.concatenate(
                            (np.zeros(-diff_start), bottom))
                    if diff_end > 0:
                        bottom = np.concatenate((bottom, np.zeros(diff_end)))
                    else:
                        bottom = bottom[:diff_end-1]
                    b_len, h_len = len(bottom), len(fr)
                    if  b_len > h_len:
                        bottom = bottom[:h_len]
                    elif b_len < h_len:
                        bottom = np.concatenate(
                            (bottom, np.zeros(h_len-b_len)))
                else:
                    bottom = bottom[:-1]

                ax2.fill_between(fr_times, fr + bottom, bottom, color=color)
                lines.extend(ax2.plot(fr_times, fr + bottom, ls="", marker=""))
            else:
                ax2.fill_between(fr_times, fr, 0., color=color)
                lines.extend(ax2.plot(fr_times, fr, ls="", marker=""))

            ax2.set_ylabel("Rate (Hz)")
            ax2.set_xlabel(xlabel)
            ax2.set_xlim(ax1.get_xlim())
            _second_axis(sort, sort_attribute, ax1)
        else:
            if axis is not None:
                ax = axis
            else:
                num_axes = len(fig.axes)
                for i, old_ax in enumerate(fig.axes):
                    old_ax.change_geometry(num_axes + 1, 1, i+1)
                ax = fig.add_subplot(num_axes + 1, 1, num_axes + 1)

            if limits is not None:
                start, stop = limits

                keep    = (times >= start)&(times <= stop)
                times   = times[keep]
                senders = senders[keep]

            if network is not None:
                pop = network.population
                colors = palette_discrete(np.linspace(0, 1, len(pop)))
                mm = itertools.cycle(markers)
                for m, (k, v), c in zip(mm, pop.items(), colors):
                    keep = np.where(
                        np.in1d(senders, network.nest_gids[v.ids]))[0]
                    if len(keep):
                        if label is None:
                            mpl_kwargs['label'] = k
                        lines.extend(ax.plot(
                            times[keep], senders[keep], c=c, marker=m,
                            ls='None', mec='k', mew=0.5, ms=4, **mpl_kwargs))
            else:
                lines.extend(ax.plot(
                    times, senders, c=color, marker="o", linestyle='None',
                    mec="k", mew=0.5, ms=4, **mpl_kwargs))

            ax.set_ylabel(ylabel)
            ax.set_xlabel(xlabel)

            if limits is not None:
                ax.set_xlim(limits)
            else:
                _set_ax_lims(ax, np.max(times), np.min(times), np.max(senders),
                             np.min(senders))

            if label is not None:
                ax.legend(bbox_to_anchor=(1.1, 1.2))
            _second_axis(sort, sort_attribute, ax)

        fig.suptitle(title)

        if show:
            plt.show()
    else:
        _log_message(logger, "WARNING",
                     "No activity was detected during the simulation.")

    return lines


#-----------------------------------------------------------------------------
# Tools
#------------------------
#

def _fill_between_steps(x, y1, y2=0, h_align='mid'):
    '''
    Fills a hole in matplotlib: fill_between for step plots.

    Parameters :
    ------------
    x : array-like
        Array/vector of index values. These are assumed to be equally-spaced.
        If not, the result will probably look weird...
    y1 : array-like
        Array/vector of values to be filled under.
    y2 : array-Like
        Array/vector or bottom values for filled area. Default is 0.
    '''
    # First, duplicate the x values
    xx = np.repeat(x,2)
    # Now: the average x binwidth
    xstep = np.repeat((x[1:] - x[:-1]), 2)
    xstep = np.concatenate(([xstep[0]], xstep, [xstep[-1]]))
    # Now: add one step at end of row.
    #~ xx = np.append(xx, xx.max() + xstep[-1])

    # Make it possible to change step alignment.
    if h_align == 'mid':
        xx -= xstep / 2.
    elif h_align == 'right':
        xx -= xstep

    # Also, duplicate each y coordinate in both arrays
    y1 = np.repeat(y1,2)#[:-1]
    if type(y2) == np.ndarray:
        y2 = np.repeat(y2,2)#[:-1]

    return xx, y1, y2


def _moving_average (values, window):
    weights = np.repeat(1.0, window)/window
    sma = np.convolve(values, weights, 'same')
    return sma


def _second_axis(sort, sort_attribute, ax):
    import matplotlib.pyplot as plt
    if sort is not None:
        fig = ax.get_figure()
        twin = None
        for axis in fig.axes:
            if axis.get_ylabel() == sort:
                twin = axis
                break
        if twin is None:
            asort = np.argsort(sort_attribute)
            twin = ax.twinx()
            twin.grid(False)
            twin.set_ylabel(sort)
            plt.draw()
            old_ticks = ax.get_yticks()
            twin.set_yticks(old_ticks)
            twin.set_ylim(ax.get_ylim())
            labels = ['' for _ in range(len(old_ticks))]
            idx_max = len(sort_attribute) - 1
            for i, t in enumerate(old_ticks):
                if t >= 0:
                    idx = min(int(t), idx_max)
                    labels[i] = _sci_format(sort_attribute[asort[idx]])
            twin.set_yticklabels(labels)


def _sci_format(n):
    label = ''
    if np.abs(n) < 0.01 or np.abs(n) >= 1000:
        a = '{:.1E}'.format(n)
        label = '$' + a.split('E')[0].rstrip('0').rstrip('.') + '\\cdot 10^{'
        exponent = a.split('E')[1].lstrip('0')
        if exponent[0] == '-':
            exponent = exponent[0] + exponent[1:].lstrip('0')
        elif exponent[0] == '+':
            exponent = exponent[1:].lstrip('0')
        label += exponent + '}$'
    elif np.abs(n) >= 100:
       label = '{:.0f}'.format(n)
    elif np.abs(n) >= 10:
       label = '{:.1f}'.format(n)
    else:
       label = '{:.2f}'.format(n)
    return label

def _make_patch_spines_invisible(ax):
    ax.set_frame_on(True)
    ax.patch.set_visible(False)
    for sp in ax.spines.values():
        sp.set_visible(False)



#-*- coding:utf-8 -*-
#
# lib/sorting.py
#
# This file is part of the NNGT project, a graph-library for standardized and
# and reproducible graph analysis: generate and analyze networks with your
# favorite graph library (graph-tool/igraph/networkx) on any platform, without
# any change to your code.
# Copyright (C) 2015-2021 Tanguy Fardet
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

""" Sorting tools """

import numpy as np

from .errors import InvalidArgument
from .test_functions import nonstring_container, is_integer


def find_idx_nearest(array, values):
    '''
    Find the indices of the nearest elements of `values` in a sorted `array`.

    .. warning::
        Both ``array`` and ``values`` should be `numpy.array` objects and
        `array` MUST be sorted in increasing order.

    Parameters
    ----------
    array : reference list or np.ndarray
    values : double, list or array of values to find in `array`

    Returns
    -------
    idx : int or array representing the index of the closest value in `array`
    '''
    idx = np.searchsorted(array, values, side="left") # get the interval
    # return the index of the closest
    if isinstance(values, np.float) or is_integer(values):
        if idx == len(array):
            return idx-1
        else:
            idx -= (np.abs(values-array[idx-1]) < np.abs(values-array[idx]))
            return idx
    else:
        # find where it is idx_max+1
        overflow = (idx == len(array))
        idx[overflow] -= 1
        # for the others, find the nearest
        tmp = idx[~overflow]
        idx[~overflow] = tmp - (np.abs(values[~overflow] - array[tmp-1])
                                < np.abs(values[~overflow] - array[tmp]))
        return idx


def _sort_neurons(sort, gids, network, data=None, return_attr=False):
    '''
    Sort the neurons according to the `sort` property.

    If `sort` is "firing_rate" or "B2", then data must contain the `senders`
    and `times` list given by a NEST ``spike_recorder``.

    Parameters
    ----------
    sort : str or array
        Sorting method or indices
    gids : array-like
        NEST gids
    network : the network
    data : numpy.array of shape (N, 2)
        Senders on column 1, times on column 2.

    Returns
    -------
    For N neurons, labeled from ``GID_MIN`` to ``GID_MAX``, returns a`sorting`
    array of size ``GID_MAX``, where ``sorting[gids]`` gives the sorted ids of
    the neurons, i.e. an integer between 1 and N.
    '''
    from nngt.analysis import node_attributes, get_b2
    from nngt.simulation.nest_utils import nest_version

    min_nest_gid = network.nest_gids.min()
    max_nest_gid = network.nest_gids.max()

    sorting = np.zeros(max_nest_gid + 1)

    attribute = None
    sorted_ids = None

    if isinstance(sort, str):
        if sort == "firing_rate":
            # compute number of spikes per neuron
            spikes = np.bincount(data[:, 0].astype(int))

            # shift from 1 in NEST 3+
            max_entry = max_nest_gid + 1 if nest_version == 3 else max_nest_gid

            if spikes.shape[0] < max_entry: # one entry per neuron
                spikes.resize(max_entry)

            # sort them (neuron with least spikes arrives at min_nest_gid)
            sorted_ids = np.argsort(spikes)[min_nest_gid:] - min_nest_gid

            # get attribute
            if len(data[:, 0]):
                idx_min = int(np.min(data[:, 0]))
                attribute = spikes[idx_min:] \
                            / (np.max(data[:, 1]) - np.min(data[:, 1]))
            else:
                attribute = []
        elif sort.lower() == "b2":
            attribute = get_b2(network, data=data, nodes=gids)
            sorted_ids = np.argsort(attribute)
            # check for non-spiking neurons
            num_b2 = attribute.shape[0]
            if num_b2 < network.node_nb():
                spikes = np.bincount(data[:, 0])
                non_spiking = np.where(spikes[min_nest_gid:] == 0)[0]
                sorted_ids.resize(network.node_nb())
                for i, n in enumerate(non_spiking):
                    sorted_ids[sorted_ids >= n] += 1
                    sorted_ids[num_b2 + i] = n
        elif sort == "space":
            xs, ys = network.get_positions().T
            x_min, x_max = np.min(xs), np.max(xs)
            y_min, y_max = np.min(ys), np.max(ys)

            num_boxes = 10

            xbins  = np.linspace(x_min, x_max, num_boxes + 1)
            ybins  = np.linspace(y_min, y_max, num_boxes + 1)
            xboxes = np.digitize(xs, xbins)
            yboxes = np.digitize(ys, ybins)

            attribute  = xboxes*num_boxes + yboxes
            sorted_ids = np.argsort(attribute)
        elif sort == "x":
            xs, _ = network.get_positions().T
            x_min, x_max = np.min(xs), np.max(xs)

            num_boxes = 10

            xbins  = np.linspace(x_min, x_max, num_boxes + 1)
            attribute = np.digitize(xs, xbins)
            sorted_ids = np.argsort(attribute)
        elif sort == "y":
            _, ys = network.get_positions().T
            y_min, y_max = np.min(ys), np.max(ys)

            num_boxes = 10

            ybins  = np.linspace(y_min, y_max, num_boxes + 1)
            attribute = np.digitize(ys, ybins)
            sorted_ids = np.argsort(attribute)
        else:
            attribute = node_attributes(network, sort)
            sorted_ids = np.argsort(attribute)
    else:
        sorted_ids = np.argsort(sort)
        attribute  = sort[sorted_ids]

    if network.is_network():
        num_sorted = 1
        for group in network.population.values():
            gids = network.nest_gids[group.ids]
            order = np.argsort(np.argsort(np.argsort(sorted_ids)[group.ids]))
            sorting[gids] = num_sorted + order
            num_sorted += len(group.ids)

    if return_attr:
        return sorting.astype(int), attribute

    return sorting.astype(int)


def _sort_groups(pop):
    '''
    Sort the groups of a NeuralPop by decreasing size.
    '''
    names, groups = [], []
    for name, group in pop.items():
        names.append(name)
        groups.append(group)
    sizes = [len(g.ids) for g in groups]
    order = np.argsort(sizes)[::-1]
    return [names[i] for i in order], [groups[i] for i in order]



#!/usr/bin/env python
#-*- coding:utf-8 -*-

# test_random.py

# This file is part of the NNGT module
# Distributed as a free software, in the hope that it will be useful, under the
# terms of the GNU General Public License.

"""
Test the random seeding and generation
"""

import os

import numpy as np
import pytest

import nngt
import nngt.generation as ng


if os.environ.get("MPI"):
    nngt.set_config("mpi", True)

if os.environ.get("OMP"):
    nngt.set_config("omp", int(os.environ["OMP"]))


def get_num_seeds():
    if nngt.get_config("multithreading"):
        return nngt.get_config("omp")

    return nngt.num_mpi_processes()


# ----- #
# Tests #
# ----- #

def test_random_seeded():
    num_seeds = get_num_seeds()

    # test equality of graph generated with same seeds
    nngt.seed(msd=0, seeds=[i for i in range(1, num_seeds + 1)])
    g1 = ng.gaussian_degree(10, 1, nodes=100)

    nngt.seed(msd=0, seeds=[i for i in range(1, num_seeds + 1)])
    g2 = ng.gaussian_degree(10, 1, nodes=100)

    assert np.all(g1.edges_array == g2.edges_array)

    # check that subsequent graphs are different
    g3 = ng.gaussian_degree(10, 1, nodes=100)

    # with mpi onnon-distributed backends, test only on master process
    if nngt.get_config("backend") == "nngt" or nngt.on_master_process():
        if g3.edge_nb() == g2.edge_nb():
            assert np.any(g2.edges_array != g3.edges_array)


def test_random_unseeded():
    num_seeds = get_num_seeds()

    nngt.seed(msd=42)

    # check that seeds generated by first call indeed allow to reproduce the
    # graph in later calls
    g1 = ng.gaussian_degree(10, 1, nodes=100)

    seeds = nngt.get_config("seeds")

    nngt.seed(msd=42, seeds=seeds)
    g2 = ng.gaussian_degree(10, 1, nodes=100)

    assert np.all(g1.edges_array == g2.edges_array)


# --------- #
# Run tests #
# --------- #

if __name__ == "__main__":
    test_random_seeded()
    test_random_unseeded()



#!/usr/bin/env python
#-*- coding:utf-8 -*-
#
# This file is part of the PyNCulture project, which aims at providing tools to
# easily generate complex neuronal cultures.
# Copyright (C) 2017 SENeC Initiative
# 
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# 
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

'''
Geometry utility functions.
'''


_di_mag = {
    'um': 1e-6,
    'mm': 1e-3,
    'cm': 1e-2,
    'dm': 0.1,
    'm': 1.
}


def conversion_magnitude(source_unit, target_unit):
    '''
    Returns the magnitude necessary to convert from values in `source_unit` to
    values in `target_unit`.
    '''
    return _di_mag[source_unit] / _di_mag[target_unit]



#-*- coding:utf-8 -*-
#
# core/graph_interface.py
#
# This file is part of the NNGT project, a graph-library for standardized and
# and reproducible graph analysis: generate and analyze networks with your
# favorite graph library (graph-tool/igraph/networkx) on any platform, without
# any change to your code.
# Copyright (C) 2015-2021 Tanguy Fardet
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

""" Base classes to interface the various libraries' graphs """

from abc import abstractmethod, abstractproperty
from collections import OrderedDict
from copy import deepcopy
from weakref import ref

import logging

import numpy as np
from scipy.sparse import csr_matrix

import nngt
from nngt.lib import InvalidArgument, BWEIGHT, nonstring_container, is_integer
from nngt.lib.graph_helpers import _get_edge_attr, _get_syn_param
from nngt.lib.converters import _np_dtype, _to_np_array
from nngt.lib.logger import _log_message


logger = logging.getLogger(__name__)


# ---------------------------------- #
# Library-dependent graph properties #
# ---------------------------------- #

class BaseProperty(dict):

    def __init__(self, parent):
        self.parent = ref(parent)
        self._num_values_set = {}

    def value_type(self, key=None):
        if key is not None:
            return super(type(self), self).__getitem__(key)
        else:
            return {k: super(type(self), self).__getitem__(k) for k in self}

    # redefine dict values/items to use the __getitem__ that will be
    # overwritten by the child classes

    def values(self):
        return [self[k] for k in self]

    def items(self):
        return [(k, self[k]) for k in self]


# -------------- #
# GraphInterface #
# -------------- #

class GraphInterface:

    #------------------------------------------------------------------#
    # Class attributes

    _nattr_class = None
    _eattr_class = None

    #------------------------------------------------------------------#
    # Shared properties methods

    @property
    @abstractmethod
    def edges_array(self):
        pass

    def remove_edge(self, edge):
        raise NotImplementedError(
            "This function has been removed because it makes using edge "
            "properties too complicated.")

    def remove_vertex(self, node, fast=False):
        raise NotImplementedError(
            "This function has been removed because it makes using node"
            "properties too complicated.")

    def remove_node(self, node, fast=False):
        raise NotImplementedError(
            "This function has been removed because it makes using node"
            "properties too complicated.")

    #------------------------------------------------------------------#
    # Properties and methods to implement

    #~ @abstractmethod
    #~ def inhibitory_subgraph(self):
        #~ pass
    #~
    #~ @abstractmethod
    #~ def excitatory_subgraph(self, n=1, neuron_type=1):
        #~ pass

    @abstractmethod
    def edge_id(self, edge):
        pass

    @abstractmethod
    def has_edge(self, edge):
        pass

    @abstractmethod
    def new_node(self, n=1, neuron_type=1, attributes=None):
        pass

    @abstractmethod
    def delete_nodes(self, nodes):
        pass

    @abstractmethod
    def new_edge(self, source, target, attributes=None, ignore=False):
        pass

    @abstractmethod
    def new_edges(self, edge_list, attributes=None, check_duplicates=False,
                  check_self_loops=True, check_existing=True,
                  ignore_invalid=False):
        pass

    @abstractmethod
    def delete_edges(self, edges):
        pass

    @abstractmethod
    def node_nb(self):
        pass

    @abstractmethod
    def edge_nb(self):
        pass

    @abstractmethod
    def get_degrees(self, mode="total", nodes=None, weights=False):
        pass

    @abstractmethod
    def neighbours(self, node, mode="all"):
        pass

    @abstractmethod
    def clear_all_edges(self):
        pass

    @abstractmethod
    def _from_library_graph(self, graph, copy=True):
        pass

    def _attr_new_edges(self, edge_list, attributes=None):
        ''' Generate attributes for newly created edges. '''
        num_edges = len(edge_list)

        if num_edges:
            attributes = {} if attributes is None else attributes
            specials = ("weight", "delay", 'distance')

            for k in attributes.keys():
                if k not in self.edge_attributes and k in specials:
                    self._eattr.new_attribute(name=k, value_type="double")

            # take care of classic attributes
            bio_weights = False
            bio_delays = False

            # distance must come first
            if self.is_spatial() or "distance" in attributes:
                prop = attributes.get("distance", None)
                values = _get_edge_attr(
                    self, edge_list, 'distance', prop, last_edges=True)
                self._eattr.set_attribute(
                    "distance", values, edges=edge_list, last_edges=True)

            # first check for potential syn_spec if Network
            if self.is_network():
                for syn_param in self.population.syn_spec.values():
                    bio_weights += ("weight" in syn_param)
                    bio_delays  += ("delay" in syn_param)

            # then weights
            if bio_weights:
                syn_spec = self.population.syn_spec

                mat = csr_matrix(
                    (np.repeat(1., num_edges),
                    (edge_list[:, 0], edge_list[:, 1])),
                    (self.population.size, self.population.size))

                for name1, g1 in self.population.items():
                    for name2, g2 in self.population.items():
                        src_slice = slice(g1.ids[0], g1.ids[-1]+1)
                        tgt_slice = slice(g2.ids[0], g2.ids[-1]+1)

                        e12 = mat[src_slice, tgt_slice].nonzero()
    
                        syn_prop = _get_syn_param(
                            name1, g1, name2, g2, syn_spec, "weight")
                        syn_prop = 1. if syn_prop is None else syn_prop

                        if isinstance(syn_prop, dict):
                            # through set_weights for dict
                            distrib = syn_prop["distribution"]
                            del syn_prop["distribution"]
                            self.set_weights(elist=e12, distribution=distrib,
                                             parameters=syn_prop)
                        elif nonstring_container(syn_prop):
                            # otherwise direct attribute set
                            self.set_edge_attribute(
                                "weight", values=syn_prop, value_type="double",
                                edges=edge_list)
                        else:
                            self.set_edge_attribute(
                                "weight", val=syn_prop, value_type="double",
                                edges=edge_list)
            elif self.is_weighted() or "weight" in attributes:
                values = _get_edge_attr(
                    self, edge_list, 'weight', attributes.get("weight", None),
                    last_edges=True)

                self._eattr.set_attribute(
                    "weight", values, edges=edge_list, last_edges=True)

            # then delay
            if self.is_network() or "delay" in self.edge_attributes:
                prop = attributes.get("delay", None)

                values = _get_edge_attr(
                    self, edge_list, 'delay', prop, last_edges=True)

                self._eattr.set_attribute(
                    "delay", values, edges=edge_list, last_edges=True)

            for k in attributes.keys():
                if k not in specials:
                    if k in self.edge_attributes:
                        values = _get_edge_attr(
                            self, edge_list, k, attributes[k], last_edges=True)

                        self._eattr.set_attribute(k, values, edges=edge_list,
                                                  last_edges=True)
                    else:
                        raise RuntimeError("Unknown attribute: '" + k + "'.")

            # take care of potential new attributes
            if "names" in attributes:
                num_attr = len(attributes["names"])

                for i in range(num_attr):
                    v = attributes["values"][i]

                    if not nonstring_container(v):
                        v = np.repeat(v, self.edge_nb())

                    self._eattr.new_attribute(attributes["names"][i],
                                              attributes["types"][i], values=v)





#-*- coding:utf-8 -*-
#
# geospatial/_cartopy_ne.py
#
# This file is part of the NNGT project, a graph-library for standardized and
# and reproducible graph analysis: generate and analyze networks with your
# favorite graph library (graph-tool/igraph/networkx) on any platform, without
# any change to your code.
# Copyright (C) 2015-2021 Tanguy Fardet
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

import io
import os

import cartopy
from cartopy.io import Downloader


"""
Temporary fix for Cartopy loading NaturalEarth data until the following PR
is merge: https://github.com/SciTools/cartopy/pull/1745

@todo: remove.
It is merged since Cartopy 0.19 but I'm waiting a bit to require it.
"""

class NEShpDownloader(Downloader):
    '''
    Specialise :class:`cartopy.io.Downloader` to download the zipped
    Natural Earth shapefiles and extract them to the defined location
    (typically user configurable).

    The keys which should be passed through when using the ``format_dict``
    are typically ``category``, ``resolution`` and ``name``.
    '''
    FORMAT_KEYS = ('config', 'resolution', 'category', 'name')

    # Define the NaturalEarth URL template. The natural earth website
    # returns a 302 status if accessing directly, so we use the naciscdn
    # URL directly.
    _NE_URL_TEMPLATE = ('https://naciscdn.org/naturalearth/{resolution}'
                        '/{category}/ne_{resolution}_{name}.zip')

    def __init__(self, url_template=_NE_URL_TEMPLATE,
                 target_path_template=None, pre_downloaded_path_template=''):
        ''' adds some NE defaults to the __init__ of a Downloader'''
        super().__init__(url_template, target_path_template,
                         pre_downloaded_path_template)

    def zip_file_contents(self, format_dict):
        '''
        Return a generator of the filenames to be found in the downloaded
        natural earth zip file.

        '''
        for ext in ['.shp', '.dbf', '.shx', '.prj', '.cpg']:
            yield ('ne_{resolution}_{name}'
                   '{extension}'.format(extension=ext, **format_dict))

    def acquire_resource(self, target_path, format_dict):
        '''
        Download the zip file and extracts the files listed in
        :meth:`zip_file_contents` to the target path.

        '''
        from zipfile import ZipFile

        target_dir = os.path.dirname(target_path)
        if not os.path.isdir(target_dir):
            os.makedirs(target_dir)

        url = self.url(format_dict)

        shapefile_online = self._urlopen(url)

        zfh = ZipFile(io.BytesIO(shapefile_online.read()), 'r')

        for member_path in self.zip_file_contents(format_dict):
            ext = os.path.splitext(member_path)[1]
            target = os.path.splitext(target_path)[0] + ext
            member = zfh.getinfo(member_path.replace(os.sep, '/'))
            with open(target, 'wb') as fh:
                fh.write(zfh.open(member).read())

        shapefile_online.close()
        zfh.close()

        return target_path

    @staticmethod
    def default_downloader():
        '''
        Return a generic, standard, NEShpDownloader instance.

        Typically, a user will not need to call this staticmethod.

        To find the path template of the NEShpDownloader:

            >>> ne_dnldr = NEShpDownloader.default_downloader()
            >>> print(ne_dnldr.target_path_template)
            {config[data_dir]}/shapefiles/natural_earth/{category}/\
ne_{resolution}_{name}.shp
        '''
        default_spec = ('shapefiles', 'natural_earth', '{category}',
                        'ne_{resolution}_{name}.shp')
        ne_path_template = os.path.join('{config[data_dir]}', *default_spec)
        pre_path_template = os.path.join('{config[pre_existing_data_dir]}',
                                         *default_spec)
        return NEShpDownloader(target_path_template=ne_path_template,
                               pre_downloaded_path_template=pre_path_template)

    
# add a generic Natural Earth shapefile downloader to the cartopy config 
# dictionary's 'downloaders' section.
_ne_key = ('shapefiles', 'natural_earth')
cartopy.config['downloaders'].setdefault(_ne_key,
                                         NEShpDownloader.default_downloader())


def natural_earth(resolution='110m', category='physical', name='coastline'):
    '''
    Return the path to the requested natural earth shapefile,
    downloading and unzipping if necessary.

    To identify valid components for this function, either browse
    NaturalEarthData.com, or if you know what you are looking for, go to
    https://github.com/nvkelso/natural-earth-vector/tree/master/zips to
    see the actual files which will be downloaded.

    Note
    ----
        Some of the Natural Earth shapefiles have special features which are
        described in the name. For example, the 110m resolution
        "admin_0_countries" data also has a sibling shapefile called
        "admin_0_countries_lakes" which excludes lakes in the country
        outlines. For details of what is available refer to the Natural Earth
        website, and look at the "download" link target to identify
        appropriate names.
    '''
    # get hold of the Downloader (typically a NEShpDownloader instance)
    # which we can then simply call its path method to get the appropriate
    # shapefile (it will download if necessary)
    ne_downloader = NEShpDownloader.default_downloader()
    format_dict = {'config': cartopy.config, 'category': category,
                   'name': name, 'resolution': resolution}
    return ne_downloader.path(format_dict)
                                         



#!/usr/bin/env python
#-*- coding:utf-8 -*-
#
# This file is part of the PyNCulture project, which aims at providing tools to
# easily generate complex neuronal cultures.
# Copyright (C) 2017 SENeC Initiative
# 
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# 
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# These tools are inspired by Sean Gillies' `descartes` library, that you can
# find here: https://pypi.python.org/pypi/descartes. They are released under
# a BSD license.

""" Plotting functions for shapely objects """

import numpy as np

from matplotlib.patches import PathPatch
from matplotlib.path import Path


def plot_shape(shape, axis=None, m='', mc="#999999", fc="#8888ff",
               ec="#444444", alpha=0.5, brightness="height", show_contour=True,
               show=True, **kwargs):
    '''
    Plot a shape (you should set the `axis` aspect to 1 to respect the
    proportions).

    Parameters
    ----------
    shape : :class:`Shape`
        Shape to plot.
    axis : :class:`matplotlib.axes.Axes` instance, optional (default: None)
        Axis on which the shape should be plotted. By default, a new figure
        is created.
    m : str, optional (default: invisible)
        Marker to plot the shape's vertices, matplotlib syntax.
    mc : str, optional (default: "#999999")
        Color of the markers.
    fc : str, optional (default: "#8888ff")
        Color of the shape's interior.
    ec : str, optional (default: "#444444")
        Color of the shape's edges.
    alpha : float, optional (default: 0.5)
        Opacity of the shape's interior.
    brightness : str, optional (default: height)
        Show how different other areas are from the 'default_area' (lower
        values are darker, higher values are lighter).
        Difference can concern the 'height', or any of the `properties` of the
        :class:`Area` objects.
    show_contour : bool, optional (default: True)
        Whether the shapes should be drawn with a contour.
    show : bool, optional (default: True)
        Whether the plot should be displayed immediately.
    **kwargs: keywords arguments for :class:`matplotlib.patches.PathPatch`
    '''
    # import
    import matplotlib.pyplot as plt
    MultiPolygon = None
    try:
        from shapely.geometry import (MultiPolygon, Polygon, LineString,
                                      MultiLineString)
    except ImportError:
        pass

    if axis is None:
        fig, axis = plt.subplots()

    zorder = kwargs.get("zorder", 0)
    if "zorder" in kwargs:
        del kwargs["zorder"]

    # plot the main shape
    if isinstance(shape, MultiPolygon):
        for p in shape:
            plot_shape(p, axis=axis, m=m, mc=mc, fc=fc, ec=ec, alpha=alpha,
                       brightness=brightness, show=False,
                       show_contour=show_contour, **kwargs)
    elif isinstance(shape, Polygon) and shape.exterior.coords:
        if show_contour:
            _plot_coords(axis, shape.exterior, m, mc, ec)
            for path in shape.interiors:
                _plot_coords(axis, path.coords, m, mc, ec)
        patch = _make_patch(shape, color=fc, alpha=alpha, zorder=zorder,
                            **kwargs)
        axis.add_patch(patch)

        # take care of the areas
        if hasattr(shape, "areas"):
            def_area = shape.areas["default_area"]

            # get the highest and lowest properties
            mean = _get_prop(def_area, brightness)
            low, high = np.inf, -np.inf

            for name, area in shape.areas.items():
                if name != "default_area":
                    prop = _get_prop(area, brightness)
                    if prop < low:
                        low = prop
                    if prop > high:
                        high = prop

            # plot the areas
            for name, area in shape.areas.items():
                if name != "default_area":
                    prop         = _get_prop(area, brightness)
                    color        = fc
                    local_alpha  = 0
                    if prop < mean:
                        color       = "black"
                        local_alpha = alpha * (prop - mean) / (low - mean)
                    elif prop > mean:
                        color       = "white"
                        local_alpha = alpha * (prop - mean) / (high - mean)
                    # contour
                    _plot_coords(axis, area.exterior, m, mc, ec)
                    for path in shape.interiors:
                        _plot_coords(axis, path.coords, m, mc, ec)
                    # content
                    patch = _make_patch(
                        area, color=color, alpha=local_alpha, zorder=zorder,
                        **kwargs)
                    axis.add_patch(patch)
    elif isinstance(shape, (LineString, MultiLineString)):
        lines = [shape] if isinstance(shape, LineString) else shape.geoms
        for line in lines:
            _plot_coords(axis, line.coords, m, mc, ec)

    axis.set_aspect(1)

    if show:
        plt.show()


def _make_patch(shape, **kwargs):
    '''
    Construct a matplotlib patch from a geometric object

    Parameters
    ----------
    shape: :class:`NetGrowth.geometry.Shape`
        may be a Shapely or GeoJSON-like object with or without holes.
    kwargs: keywords arguments for :class:`matplotlib.patches.PathPatch`

    Returns
    -------
    an instance of :class:`matplotlib.patches.PathPatch`.

    Example
    -------
    (using Shapely Point and a matplotlib axes):

      >>> b = Point(0, 0).buffer(1.0)
      >>> patch = PolygonPatch(b, fc='blue', ec='blue', alpha=0.5)
      >>> axis.add_patch(patch)

    Modified from `descartes` by Sean Gillies (BSD license).
    '''
    vertices = np.concatenate(
        [np.asarray(shape.exterior.coords)[:, :2]] +
        [np.asarray(h.coords)[:, :2] for h in shape.interiors])
    instructions = np.concatenate(
        [_path_instructions(shape.exterior)] +
        [_path_instructions(h) for h in shape.interiors])

    path = Path(vertices, instructions)
    return PathPatch(path, **kwargs)


def _path_instructions(ob):
    '''
    Give instructions to build path from vertices.
    '''
    # The codes will be all "LINETO" commands, except for "MOVETO"s at the
    # beginning of each subpath
    n = len(ob.coords)
    vals = np.ones(n, dtype=Path.code_type) * Path.LINETO
    vals[0] = Path.MOVETO
    return vals


def _plot_coords(ax, ob, m, mc, ec):
    if hasattr(ob, 'coords') and isinstance(ob.coords, list):
        for coord in ob.coords:
            x, y = ob.xy
            ax.plot(x, y, m, ls='-', c=ec, markerfacecolor=mc, zorder=1)
    else:
        x, y = ob.xy
        ax.plot(x, y, m, ls='-', c=ec, markerfacecolor=mc, zorder=1)


def _get_prop(area, brightness):
    is_height = (brightness == "height")
    return area.height if is_height else area.properties[brightness]



#-*- coding:utf-8 -*-
#
# simulation/nest_graph.py
#
# This file is part of the NNGT project, a graph-library for standardized and
# and reproducible graph analysis: generate and analyze networks with your
# favorite graph library (graph-tool/igraph/networkx) on any platform, without
# any change to your code.
# Copyright (C) 2015-2021 Tanguy Fardet
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

import nest
import numpy as np
import scipy.sparse as ssp
from scipy.optimize import root
from scipy.signal import argrelmax, argrelmin

import nngt
from nngt.lib import InvalidArgument, nonstring_container, WEIGHT, DELAY
from nngt.lib.sorting import _sort_groups
from nngt.lib.test_functions import mpi_checker
from nngt.lib.graph_helpers import _get_syn_param
from .nest_utils import nest_version, _get_nest_gids


__all__ = [
    'make_nest_network',
    'get_nest_adjacency',
    'reproducible_weights'
]


# -------- #
# Topology #
# -------- #

@mpi_checker()
def make_nest_network(network, send_only=None, weights=True):
    '''
    Create a new network which will be filled with neurons and
    connector objects to reproduce the topology from the initial network.

    .. versionchanged:: 0.8
        Added `send_only` parameter.

    Parameters
    ----------
    network: :class:`nngt.Network` or :class:`nngt.SpatialNetwork`
        the network we want to reproduce in NEST.
    send_only : int, str, or list of str, optional (default: None)
        Restrict the nodes that are created in NEST to either inhibitory or
        excitatory neurons `send_only` :math:`\in \{ 1, -1\}` to a group or a
        list of groups.
    weights : bool or str, optional (default: binary edges)
        Whether edge weights should be considered; if ``None`` or ``False``
        then use binary edges; if ``True``, uses the 'weight' edge attribute,
        otherwise uses any valid edge attribute required.

    Returns
    -------
    gids : tuple or NodeCollection (nodes in NEST)
        GIDs of the neurons in the NEST network.
    '''
    gids = _get_nest_gids([])
    pop  = network.population

    send = list(network.population.keys())

    if send_only in (-1, 1):
        send = [g for g in send if pop[g].neuron_type == send_only]
    elif isinstance(send_only, str):
        send = [pop[send_only]]
    elif nonstring_container(send_only):
        send = [g for g in send_only]

    send = [g for g in send if pop[g].ids]

    # link NEST Gids to nngt.Network ids as neurons are created
    num_neurons  = network.node_nb()
    ia_nngt_ids  = np.full(num_neurons, -1, dtype=int)
    ia_nest_gids = np.full(num_neurons, -1, dtype=int)
    ia_nngt_nest = np.full(num_neurons, -1, dtype=int)
    current_size = 0

    for name in send:
        group = pop[name]
        group_size = len(group.ids)

        if group_size:
            ia_nngt_ids[current_size:current_size + group_size] = group.ids

            # clean up neuron_param dict, separate scalar and non-scalar
            defaults     = nest.GetDefaults(group.neuron_model)
            scalar_param = {}
            ns_param     = {}

            for key, val in group.neuron_param.items():
                if key in defaults and key != "model":
                    if nonstring_container(val):
                        ns_param[key] = val
                    else:
                        scalar_param[key] = val

            # create neurons:
            gids_tmp = nest.Create(group.neuron_model, group_size,
                                   scalar_param, _warn=False)

            # set non-scalar properties
            for k, v in ns_param.items():
                nest.SetStatus(gids_tmp, k, v, _warn=False)

            # create ids
            idx_nest = ia_nngt_ids[
                np.arange(current_size, current_size + group_size)]

            ia_nest_gids[current_size:current_size + group_size] = gids_tmp
            ia_nngt_nest[idx_nest] = gids_tmp
            current_size += group_size
            gids += gids_tmp

    # conversions ids/gids
    network.nest_gids = ia_nngt_nest
    network._id_from_nest_gid = {
        gid: idx for (idx, gid) in zip(ia_nngt_ids, ia_nest_gids)
    }

    # get all properties as scipy.sparse.csr matrices
    csr_weights = network.adjacency_matrix(types=False, weights=weights)
    csr_delays  = network.adjacency_matrix(types=False, weights=DELAY)

    cspec = 'one_to_one'

    num_conn = 0

    for src_name in send:
        src_group = pop[src_name]
        syn_sign = src_group.neuron_type

        # local connectivity matrix and offset to correct neuron id
        arr_idx = np.sort(src_group.ids).astype(int)

        local_csr = csr_weights[arr_idx, :]

        assert local_csr.shape[1] == network.node_nb()

        if len(src_group.ids) > 0 and pop.syn_spec is not None:
            # check whether custom synapses should be used
            local_tgt_names = [name for name in send if pop[name].ids]

            for tgt_name in send:
                tgt_group = pop[tgt_name]
                # get list of targets for each
                arr_tgt_idx = np.sort(tgt_group.ids).astype(int)

                # get non-wero entries (global NNGT ids)
                keep = local_csr[:, arr_tgt_idx].nonzero()

                local_tgt_ids = arr_tgt_idx[keep[1]]
                local_src_ids = arr_idx[keep[0]]

                if len(local_tgt_ids) and len(local_src_ids):
                    # get the synaptic parameters
                    syn_spec = _get_syn_param(
                        src_name, src_group, tgt_name, tgt_group, pop.syn_spec)

                    # using A1 to get data from matrix
                    if weights:
                        syn_spec[WEIGHT] = syn_sign *\
                            csr_weights[local_src_ids, local_tgt_ids].A1
                    else:
                        syn_spec[WEIGHT] = np.repeat(syn_sign, len(tgt_ids))
    
                    syn_spec[DELAY] = \
                        csr_delays[local_src_ids, local_tgt_ids].A1

                    num_conn += len(local_src_ids)

                    # check backend
                    with_mpi = nngt.get_config("mpi")
                    if nngt.get_config("backend") == "nngt" and with_mpi:
                        comm = nngt.get_config("mpi_comm")
                        for i in range(comm.Get_size()):
                            sources = \
                                comm.bcast(network.nest_gids[local_src_ids], i)
                            targets = \
                                comm.bcast(network.nest_gids[local_tgt_ids], i)
                            sspec   = comm.bcast(syn_spec, i)
                            nest.Connect(sources, targets, syn_spec=sspec,
                                         conn_spec=cspec, _warn=False)
                            comm.Barrier()
                    else:
                        nest.Connect(
                            network.nest_gids[local_src_ids],
                            network.nest_gids[local_tgt_ids],
                            syn_spec=syn_spec, conn_spec=cspec, _warn=False)

        elif len(src_group.ids) > 0:
            # get NEST gids of sources and targets for each edge
            src_ids = network.nest_gids[local_csr.nonzero()[0] + min_sidx]
            tgt_ids = network.nest_gids[local_csr.nonzero()[1]]

            # prepare weights
            syn_spec = {
                WEIGHT: np.repeat(syn_sign, len(src_ids)).astype(float)
            }

            if weights not in {None, False}:
                syn_spec[WEIGHT] *= csr_weights[src_group.ids, :].data

            syn_spec[DELAY] = csr_delays[src_group.ids, :].data

            nest.Connect(src_ids, tgt_ids, syn_spec=syn_spec, conn_spec=cspec,
                         _warn=False)

    # tell the populaton that the network it describes was sent to NEST
    network.population._sent_to_nest()

    return gids


def get_nest_adjacency(id_converter=None):
    '''
    Get the adjacency matrix describing a NEST network.

    Parameters
    ----------
    id_converter : dict, optional (default: None)
        A dictionary which maps NEST gids to the desired neurons ids.

    Returns
    -------
    mat_adj : :class:`~scipy.sparse.lil_matrix`
        Adjacency matrix of the network.
    '''
    gids = (nest.GetNodes()[0] if nest_version == 2
            else np.asarray(nest.GetNodes()))

    n = len(gids)

    mat_adj = ssp.lil_matrix((n,n))

    if id_converter is None:
        id_converter = {idx: i for i, idx in enumerate(gids)}

    for i in range(n):
        src = id_converter[gids[i]]
        connections = nest.GetConnections(source=(gids[i],))
        info = nest.GetStatus(connections)
        for dic in info:
            mat_adj.rows[src].append(id_converter[dic['target']])
            mat_adj.data[src].append(dic[WEIGHT])

    return mat_adj


# ------- #
# Weights #
# ------- #

def reproducible_weights(weights, neuron_model, di_param={}, timestep=0.05,
                         simtime=50., num_bins=1000, log=False):
    '''
    Find the values of the connection weights that will give PSP responses of
    `min_weight` and `max_weight` in mV.

    Parameters
    ----------
    weights : list of floats
        Exact desired synaptic weights.
    neuron_model : string
        Name of the model used.
    di_param : dict, optional (default: {})
        Parameters of the model, default parameters if not supplied.
    timestep : float, optional (default: 0.01)
        Timestep of the simulation in ms.
    simtime : float, optional (default: 10.)
        Simulation time in ms (default: 10).
    num_bins : int, optional (default: 10000)
        Number of bins used to discretize the exact synaptic weights.
    log : bool, optional (default: False)
        Whether bins should use a logarithmic scale.

    Note
    ----
    If the parameters used are not the default ones, they MUST be provided,
    otherwise the resulting weights will likely be WRONG.
    '''
    min_weight = np.min(weights)
    max_weight = np.max(weights)

    # get corrected weights
    min_corr, max_corr = _find_extremal_weights(
        min_weight, max_weight, neuron_model, di_param, timestep=timestep,
        simtime=simtime)

    # bin them
    bins = None

    if log:
        log_min = np.log10(min_corr)
        log_max = np.log10(max_corr)
        bins = np.logspace(log_min, log_max, num_bins)
    else:
        bins = np.linspace(min_corr, max_corr, num_bins)

    binned_weights = _get_psp_list(
        bins, neuron_model, di_param, timestep, simtime)

    idx_binning = np.digitize(weights, binned_weights)

    return bins[ idx_binning ]


# ----- #
# Tools #
# ----- #

def _value_psp(weight, neuron_model, di_param, timestep, simtime):
    nest.ResetKernel(_warn=False)
    nest.SetKernelStatus({"resolution": timestep})

    # create neuron and recorder
    neuron = nest.Create(neuron_model, params=di_param, _warn=False)
    V_rest = nest.GetStatus(neuron)[0]["E_L"]
    nest.SetStatus(neuron, {"V_m": V_rest}, _warn=False)
    vm = nest.Create("voltmeter", params={"interval": timestep}, _warn=False)
    nest.Connect(vm, neuron, _warn=False)
    # send the initial spike
    sg = nest.Create(
        "spike_generator", params={'spike_times': [timestep],
                                   'spike_weights': weight},
        _warn=False)

    nest.Connect(sg, neuron, _warn=False)
    nest.Simulate(simtime)

    # get the max and its time
    dvm = nest.GetStatus(vm)[0]
    da_voltage = dvm["events"]["V_m"]
    idx = np.argmax(da_voltage)

    if idx == len(da_voltage - 1):
        raise InvalidArgument("simtime too short: PSP maximum is out of range")
    else:
        val = da_voltage[idx] - V_rest
        return val


def _find_extremal_weights(min_weight, max_weight, neuron_model, di_param={},
                           precision=0.1, timestep=0.01, simtime=10.):
    '''
    Find the values of the connection weights that will give PSP responses of
    `min_weight` and `max_weight` in mV.

    Parameters
    ----------
    min_weight : float
        Minimal weight.
    max_weight : float
        Maximal weight.
    neuron_model : string
        Name of the model used.
    di_param : dict, optional (default: {})
        Parameters of the model, default parameters if not supplied.
    precision : float, optional (default : -1.)
        Precision with which the result should be obtained. If the value is
        equal to or smaller than zero, it will default to 0.1% of the value.
    timestep : float, optional (default: 0.01)
        Timestep of the simulation in ms.
    simtime : float, optional (default: 10.)
        Simulation time in ms (default: 10).

    Note
    ----
    If the parameters used are not the default ones, they MUST be provided,
    otherwise the resulting weights will likely be WRONG.
    '''
    # define the function for root finding
    def _func_min(weight):
        val = _value_psp(weight, neuron_model, di_param, timestep, simtime)
        return val - min_weight
    def _func_max(weight):
        val = _value_psp(weight, neuron_model, di_param, timestep, simtime)
        return val - max_weight
    # @todo: find highest and lowest value that result in spike emission
    # get root
    min_w = root(_func_min, min_weight, tol=0.1*min_weight/100.).x[0]
    max_w = root(_func_max, max_weight, tol=0.1*max_weight/100.).x[0]
    return min_w, max_w


def _get_psp_list(bins, neuron_model, di_param, timestep, simtime):
    '''
    Return the list of effective weights from a list of NEST connection
    weights.
    '''
    nest.ResetKernel(_warn=False)
    nest.SetKernelStatus({"resolution": timestep})
    # create neuron and recorder
    neuron = nest.Create(neuron_model, params=di_param, _warn=False)
    vm = nest.Create("voltmeter", params={"interval": timestep}, _warn=False)
    nest.Connect(vm, neuron, _warn=False)
    # send the spikes
    times = [timestep + n*simtime for n in range(len(bins))]
    sg = nest.Create(
        "spike_generator", params={'spike_times': times,
                                   'spike_weights': bins},
        _warn=False)
    nest.Connect(sg, neuron, _warn=False)
    nest.Simulate((len(bins)+1)*simtime)
    # get the max and its time
    dvm = nest.GetStatus(vm)[0]
    da_voltage = dvm["events"]["V_m"]
    da_times = dvm["events"]["times"]
    da_max_psp = da_voltage[argrelmax(da_voltage)]
    da_min_psp = da_voltage[argrelmin(da_voltage)]
    da_max_psp -= da_min_psp
    if len(bins) != len(da_max_psp):
        raise InvalidArgument("simtime too short: all PSP maxima are not in "
                              "range.")
    else:
        return da_max_psp



#-*- coding:utf-8 -*-
#
# plot/plt_networks.py
#
# This file is part of the NNGT project, a graph-library for standardized and
# and reproducible graph analysis: generate and analyze networks with your
# favorite graph library (graph-tool/igraph/networkx) on any platform, without
# any change to your code.
# Copyright (C) 2015-2021 Tanguy Fardet
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

from itertools import cycle
from collections import defaultdict

import numpy as np

import matplotlib as mpl
from matplotlib.artist import Artist
from matplotlib.patches import FancyArrowPatch, ArrowStyle, FancyArrow, Circle
from matplotlib.patches import Arc, RegularPolygon, PathPatch
from matplotlib.cm import get_cmap
from matplotlib.collections import PatchCollection, PathCollection
from matplotlib.colors import ListedColormap, Normalize, ColorConverter
from matplotlib.markers import MarkerStyle
from matplotlib.transforms import Affine2D
from mpl_toolkits.axes_grid1 import make_axes_locatable

import nngt
from nngt.lib import POS, nonstring_container, is_integer
from .custom_plt import palette_continuous, palette_discrete, format_exponent
from .chord_diag import chord_diagram as _chord_diag
from .hive_helpers import *


'''
Network plotting
================

Implemented
-----------

Simple representation for spatial graphs, random distribution if non-spatial.
Support for edge-size (according to betweenness or synaptic weight).


Objectives
----------

Implement the spring-block minimization.

If edges have varying size, plot only those that are visible (size > min)

'''

__all__ = ["chord_diagram", "draw_network", "hive_plot", "library_draw"]


# ------- #
# Drawing #
# ------- #

def draw_network(network, nsize="total-degree", ncolor="group", nshape="o",
                 nborder_color="k", nborder_width=0.5, esize=1., ecolor="k",
                 ealpha=0.5, max_nsize=None, max_esize=2., curved_edges=False,
                 threshold=0.5, decimate_connections=None, spatial=True,
                 restrict_sources=None, restrict_targets=None,
                 restrict_nodes=None, restrict_edges=None,
                 show_environment=True, fast=False, size=(600, 600),
                 xlims=None, ylims=None, dpi=75, axis=None, colorbar=False,
                 cb_label=None, layout=None, show=False, **kwargs):
    '''
    Draw a given graph/network.

    Parameters
    ----------
    network : :class:`~nngt.Graph` or subclass
        The graph/network to plot.
    nsize : float, array of float or string, optional (default: "total-degree")
        Size of the nodes as a percentage of the canvas length. Otherwise, it
        can be a string that correlates the size to a node attribute among
        "in/out/total-degree", "in/out/total-strength", or "betweenness".
    ncolor : float, array of floats or string, optional (default: 0.5)
        Color of the nodes; if a float in [0, 1], position of the color in the
        current palette, otherwise a string that correlates the color to a node
        attribute among "in/out/total-degree", "betweenness" or "group".
    nshape : char, array of chars, or groups, optional (default: "o")
        Shape of the nodes (see `Matplotlib markers <http://matplotlib.org/api/
        markers_api.html?highlight=marker#module-matplotlib.markers>`_).
        When using groups, they must be pairwise disjoint; markers will be
        selected iteratively from the matplotlib default markers.
    nborder_color : char, float or array, optional (default: "k")
        Color of the node's border using predefined `Matplotlib colors
        <http://matplotlib.org/api/colors_api.html?highlight=color
        #module-matplotlib.colors>`_).
        or floats in [0, 1] defining the position in the palette.
    nborder_width : float or array of floats, optional (default: 0.5)
        Width of the border in percent of canvas size.
    esize : float, str, or array of floats, optional (default: 0.5)
        Width of the edges in percent of canvas length. Available string values
        are "betweenness" and "weight".
    ecolor : str, char, float or array, optional (default: "k")
        Edge color. If ecolor="groups", edges color will depend on the source
        and target groups, i.e. only edges from and toward same groups will
        have the same color.
    max_esize : float, optional (default: 5.)
        If a custom property is entered as `esize`, this normalizes the edge
        width between 0. and `max_esize`.
    threshold : float, optional (default: 0.5)
        Size under which edges are not plotted.
    decimate_connections : int, optional (default: keep all connections)
        Plot only one connection every `decimate_connections`.
        Use -1 to hide all edges.
    spatial : bool, optional (default: True)
        If True, use the neurons' positions to draw them.
    restrict_sources : str, group, or list, optional (default: all)
        Only draw edges starting from a restricted set of source nodes.
    restrict_targets : str, group, or list, optional (default: all)
        Only draw edges ending on a restricted set of target nodes.
    restrict_nodes : str, group, or list, optional (default: plot all nodes)
        Only draw a subset of nodes.
    restrict_edges : list of edges, optional (default: all)
        Only draw a subset of edges.
    show_environment : bool, optional (default: True)
        Plot the environment if the graph is spatial.
    fast : bool, optional (default: False)
        Use a faster algorithm to plot the edges. Zooming on the drawing made
        using this method leaves the size of the nodes and edges unchanged, it
        is therefore not recommended when size consistency matters, e.g. for
        some spatial representations.
    size : tuple of ints, optional (default: (600,600))
        (width, height) tuple for the canvas size (in px).
    dpi : int, optional (default: 75)
        Resolution (dot per inch).
    axis : matplotlib axis, optional (default: create new axis)
        Axis on which the network will be plotted.
    colorbar : bool, optional (default: False)
        Whether to display a colorbar for the node colors or not.
    cb_label : str, optional (default: None)
        A label for the colorbar.
    layout : str, optional (default: random or spatial positions)
        Name of a standard layout to structure the network. Available layouts
        are: "circular" or "random". If no layout is provided and the network
        is spatial, then node positions will be used by default.
    show : bool, optional (default: True)
        Display the plot immediately.
    **kwargs : dict
        Optional keyword arguments including `node_cmap` to set the
        nodes colormap (default is "magma" for continuous variables and
        "Set1" for groups) and "title" to add a title to the plot.
    '''
    import matplotlib.pyplot as plt

    # figure and axes
    size_inches = (size[0]/float(dpi), size[1]/float(dpi))

    if axis is None:
        fig = plt.figure(facecolor='white', figsize=size_inches,
                         dpi=dpi)
        axis = fig.add_subplot(111, frameon=0, aspect=1)

    # projections for geographic plots

    proj = kwargs.get("proj", None)

    kw = {} if proj is None else {"transform": proj}

    if proj is None:
        axis.set_axis_off()

    pos = None

    # arrow style
    arrowstyle = "-|>" if network.is_directed() else "-"

    # restrict sources and targets
    restrict_sources = _convert_to_nodes(restrict_sources,
                                         "restrict_sources", network)

    restrict_targets = _convert_to_nodes(restrict_targets,
                                         "restrict_targets", network)

    restrict_nodes = _convert_to_nodes(restrict_nodes,
                                       "restrict_nodes", network)

    if restrict_nodes is not None and restrict_sources is not None:
        restrict_sources = \
            set(restrict_nodes).intersection(restrict_sources)
    elif restrict_nodes is not None:
        restrict_sources = set(restrict_nodes)

    if restrict_nodes is not None and restrict_targets is not None:
        restrict_targets = \
            set(restrict_nodes).intersection(restrict_targets)
    elif restrict_nodes is not None:
        restrict_targets = set(restrict_nodes)

    # get nodes and edges
    n = network.node_nb() if restrict_nodes is None \
                          else len(restrict_nodes)

    adj_mat = network.adjacency_matrix(weights=None)

    if restrict_sources is not None:
        remove = np.array(
            [1 if node not in restrict_sources else 0
             for node in range(network.node_nb())],
            dtype=bool)
        adj_mat[remove] = 0

    if restrict_targets is not None:
        remove = np.array(
            [1 if node not in restrict_targets else 0
             for node in range(network.node_nb())],
            dtype=bool)
        adj_mat[:, remove] = 0

    edges = (np.array(adj_mat.nonzero()).T if restrict_edges is None else
             restrict_edges)

    e = len(edges)

    # compute properties
    decimate_connections = 1 if decimate_connections is None\
                           else decimate_connections

    # get node and edge shape/size properties
    simple_nodes = kwargs.get("simple_nodes", False)

    if fast:
        simple_nodes = True

    max_nsize = (20 if simple_nodes else 5) if max_nsize is None else max_nsize

    markers, nsize, esize = _node_edge_shape_size(
        network, nshape, nsize, max_nsize, esize, max_esize, restrict_nodes,
        edges, size, threshold, simple_nodes=simple_nodes)

    # node color information
    default_ncmap = (palette_discrete() if not nonstring_container(ncolor) and
                     ncolor == "group" else palette_continuous())

    nalpha = kwargs.get("nalpha", 1)

    ncmap = get_cmap(kwargs.get("node_cmap", default_ncmap))
    node_color, nticks, ntickslabels, nlabel = \
        _node_color(network, restrict_nodes, ncolor)

    if nonstring_container(ncolor):
        assert len(ncolor) == n, "For color arrays, one " +\
            "color per node is required."
        ncolor = "custom"

    c = node_color

    if not nonstring_container(nborder_color):
        nborder_color = np.repeat(nborder_color, n)

    # check edge color
    group_based = False

    default_ecmap = (palette_discrete() if not nonstring_container(ncolor) and
                     ecolor == "group" else palette_continuous())

    if ecolor == "groups" or ecolor == "group":
        if not network.is_network():
            raise TypeError(
                "The graph must be a Network to use `ecolor='groups'`.")

        group_based = True
        ecolor      = {}

        for i, src in enumerate(network.population):
            if network.population[src].ids:
                idx1 = network.population[src].ids[0]
                for j, tgt in enumerate(network.population):
                    if network.population[tgt].ids:
                        idx2 = network.population[tgt].ids[0]
                        if src == tgt:
                            ecolor[(src, tgt)] = node_color[idx1]
                        else:
                            ecolor[(src, tgt)] = \
                                np.abs(0.8*node_color[idx1]
                                       - 0.2*node_color[idx2])
    elif not nonstring_container(ecolor):
        ecolor = np.repeat(ecolor, e)

    # draw
    pos = kwargs.get("positions", None)

    spatial *= (network.is_spatial() or pos is not None)

    if nonstring_container(layout):
        assert np.shape(layout) == (n, 2), "One position per node is required."
        pos = np.asarray(layout)
    elif layout == "circular":
        pos = _circular_layout(network, nsize)
    elif pos is None and spatial:
        if show_environment:
            nngt.geometry.plot.plot_shape(network.shape, axis=axis, show=False)

        nodes = None if restrict_nodes is None else list(restrict_nodes)

        pos = network.get_positions(nodes=nodes)
    else:
        pos = np.full((n, 2), np.NaN)

        pos[:, 0] = size[0]*(np.random.uniform(size=n)-0.5)
        pos[:, 1] = size[1]*(np.random.uniform(size=n)-0.5)

    # prepare node colors
    if nonstring_container(c) and not isinstance(c[0], str):
        # make the colorbar for the nodes
        cmap = ncmap
        if colorbar:
            clist = np.unique(c, axis=0) if ncolor == "group" else None
            cnorm = None
            if ncolor.startswith("group"):
                cmap  = _discrete_cmap(len(nticks), ncmap, clist=clist)
                cnorm = Normalize(nticks[0]-0.5, nticks[-1] + 0.5)
            else:
                cnorm = Normalize(np.min(c), np.max(c))
            sm = plt.cm.ScalarMappable(cmap=cmap, norm=cnorm)
            c  = cnorm(c)
            if ncolor.startswith("group"):
                sm.set_array(nticks)
            else:
                sm.set_array(c)
            plt.subplots_adjust(right=0.95)
            divider = make_axes_locatable(axis)
            cax     = divider.append_axes("right", size="5%", pad=0.05)
            if ncolor.startswith("group"):
                cb = plt.colorbar(sm, ticks=nticks, cax=cax, shrink=0.8)
                cb.set_ticklabels(ntickslabels)
                if nlabel:
                    cb.set_label(nlabel)
            else:
                cb = plt.colorbar(sm, cax=cax, shrink=0.8)

                if cb_label is not None:
                    cb.ax.set_ylabel(cb_label)
        else:
            cmin, cmax = np.min(c), np.max(c)
            if cmin != cmax:
                c = (c - cmin)/(cmax - cmin)
        c = cmap(c)
    else:
        if not nonstring_container(c) and not isinstance(c, str):
            minc = np.min(node_color)

            c = np.array(
                [ncmap((node_color - minc)/(np.max(node_color) - minc))]*n)

    # plot nodes
    if simple_nodes:
        if nonstring_container(nshape):
            # matplotlib scatter does not support marker arrays
            if isinstance(nshape[0], nngt.Group):
                for g in nshape:
                    ids = g.ids if restrict_nodes is None \
                          else list(set(g.ids).intersection(restrict_nodes))

                    axis.scatter(pos[ids, 0], pos[ids, 1], color=c[ids],
                                 s=0.5*np.array(nsize)[ids],
                                 marker=markers[ids[0]], zorder=2,
                                 edgecolors=nborder_color,
                                 linewidths=nborder_width, alpha=nalpha)
            else:
                ids = range(network.node_nb()) if restrict_nodes is None \
                      else restrict_nodes

                for i in ids:
                    axis.plot(
                        pos[i, 0], pos[i, 1], color=c[i], ms=0.5*nsize[i],
                        marker=nshape[i], ls="", zorder=2,
                        mec=nborder_color[i], mew=nborder_width, alpha=nalpha)
        else:
            axis.scatter(pos[:, 0], pos[:, 1], color=c, s=0.5*np.array(nsize),
                         marker=nshape, zorder=2, edgecolor=nborder_color,
                         linewidths=nborder_width, alpha=nalpha)
    else:
        nodes = []

        axis.set_aspect(1.)

        if network.is_network():
            for group in network.population.values():
                idx = group.ids if restrict_nodes is None \
                      else list(set(restrict_nodes).intersection(group.ids))
                for i, fc in zip(idx, c[idx]):
                    m = MarkerStyle(markers[i]).get_path()
                    transform = Affine2D().scale(
                        0.5*nsize[i]).translate(pos[i][0], pos[i][1])
                    patch = PathPatch(m.transformed(transform), facecolor=fc,
                                      edgecolor=nborder_color[i], alpha=nalpha)
                    nodes.append(patch)
        else:
            for i, ci in enumerate(c):
                m = MarkerStyle(markers[i]).get_path()
                transform = Affine2D().scale(0.5*nsize[i]).translate(
                    pos[i][0], pos[i][1])
                patch = PathPatch(m.transformed(transform), facecolor=ci,
                                  edgecolor=nborder_color[i], alpha=nalpha)
                nodes.append(patch)

        nodes = PatchCollection(nodes, match_original=True, alpha=nalpha)
        nodes.set_zorder(2)
        axis.add_collection(nodes)

    if not show_environment or not spatial or proj is not None:
        # axis.get_data()
        _set_ax_lim(axis, pos[:, 0], pos[:, 1], xlims, ylims)

    # use quiver to draw the edges
    if e and decimate_connections != -1:
        avg_size = np.average(nsize)

        arrows = []

        if group_based:
            for src_name, src_group in network.population.items():
                for tgt_name, tgt_group in network.population.items():
                    s_ids = src_group.ids

                    if restrict_sources is not None:
                        s_ids = list(set(restrict_sources).intersection(s_ids))

                    t_ids = tgt_group.ids

                    if restrict_targets is not None:
                        t_ids = list(set(restrict_targets).intersection(t_ids))

                    if t_ids and s_ids:
                        s_min, s_max = np.min(s_ids), np.max(s_ids) + 1
                        t_min, t_max = np.min(t_ids), np.max(t_ids) + 1

                        edges = np.array(
                            adj_mat[s_min:s_max, t_min:t_max].nonzero(),
                            dtype=int)

                        edges[0, :] += s_min
                        edges[1, :] += t_min

                        strght_edges, self_loops, strght_sizes, loop_sizes = \
                            _split_edges_sizes(edges, esize,
                                               decimate_connections)

                        # plot
                        ec = default_ecmap(ecolor[(src_name, tgt_name)])

                        if fast:
                            dl       = 0.5*np.max(nsize)
                            arrow_x  = pos[strght_edges[1], 0] - \
                                       pos[strght_edges[0], 0]
                            arrow_x -= np.sign(arrow_x) * dl
                            arrow_y  = pos[strght_edges[1], 1] - \
                                       pos[strght_edges[0], 1]
                            arrow_x -= np.sign(arrow_y) * dl

                            axis.quiver(
                                pos[strght_edges[0], 0],
                                pos[strght_edges[0], 1], arrow_x,
                                arrow_y, scale_units='xy', angles='xy',
                                scale=1, alpha=ealpha, width=1.5e-3,
                                linewidths=0.5*strght_sizes, edgecolors=ec,
                                zorder=1, **kw)

                            for i, s in enumerate(self_loops):
                                es = loop_sizes[i]
                                ec = loop_colors[i]
                                cs = _connectionstyle(axis, nsize[s],
                                                      loop_sizes[i])

                                arrow = FancyArrowPatch(
                                    pos[s], pos[s], arrowstyle=arrowstyle,
                                    shrinkA=nsize[s], color=ec,
                                    linewidth=es, connectionstyle=cs, zorder=1)

                                axis.add_patch(arrow)
                        else:
                            for s, t in zip(strght_edges[0], strght_edges[1]):
                                xs, ys = pos[s, 0], pos[s, 1]
                                xt, yt = pos[t, 0], pos[t, 1]
                                dl     = 0.5*nsize[t]
                                dx     = xt-xs
                                dx -= np.sign(dx) * dl
                                dy     = yt-ys
                                dy -= np.sign(dy) * dl

                                if curved_edges:
                                    arrow = FancyArrowPatch(
                                        posA=(xs, ys), posB=(xt, yt),
                                        arrowstyle=arrowstyle,
                                        connectionstyle='arc3,rad=0.1',
                                        alpha=ealpha, fc=ec, lw=0.5)
                                    axis.add_patch(arrow)
                                elif s == t:
                                    # self-loop
                                    arrow = FancyArrowPatch(
                                        posA=(xs, ys), posB=(xt, yt),
                                        arrowstyle=arrowstyle,
                                        connectionstyle='arc3,rad=1',
                                        alpha=ealpha, fc=ec, lw=0.5)

                                    axis.add_patch(arrow)
                                else:
                                    arrows.append(FancyArrow(
                                        xs, ys, dx, dy, width=0.3*avg_size,
                                        head_length=0.7*avg_size,
                                        head_width=0.7*avg_size,
                                        length_includes_head=True,
                                        alpha=ealpha, fc=ec, lw=0.5))
        else:
            if e and decimate_connections >= 1:
                strght_colors, loop_colors = [], []

                strght_edges, self_loops, strght_sizes, loop_sizes = \
                    _split_edges_sizes(edges, esize, decimate_connections,
                    ecolor, strght_colors, loop_colors)

                # keep only desired edges
                if None not in (restrict_sources, restrict_targets):
                    new_edges = []
                    new_colors = []

                    for edge, ec in zip(strght_edges, strght_colors):
                        s, t = edge

                        if s in restrict_sources and t in restrict_targets:
                            new_edges.append(edge)
                            new_colors.append(ec)

                    strght_edges = np.array(new_edges, dtype=int)
                    strght_colors = new_colors

                    if restrict_nodes is not None:
                        nodes = list(self_loops)
                        nodes.sort()

                        new_loops = set()
                        new_colors = []

                        for i, node in enumerate(restrict_nodes):
                            strght_edges[strght_edges == node] = i

                            if node in self_loops:
                                idx = nodes.index(node)
                                new_loops.add(i)
                                new_colors.append(loop_colors[idx])

                        self_loops = new_loops
                        loop_colors = new_colors
                elif restrict_sources is not None:
                    new_edges = []
                    new_colors = []

                    for edge, ec in zip(strght_edges, strght_colors):
                        s, _ = edge

                        if s in restrict_sources:
                            new_edges.append(edge)
                            new_colors.append(ec)

                    strght_edges = np.array(new_edges, dtype=int)

                    loop_colors = [ec for ec, n in zip(loop_colors, self_loops)
                                   if n in restrict_sources]
                    self_loops  = self_loops.intersection(restrict_sources)
                elif restrict_targets is not None:
                    new_edges = []
                    new_colors = []

                    for edge, ec in zip(strght_edges, strght_colors):
                        _, t = edge

                        if t in restrict_targets:
                            new_edges.append(edge)
                            new_colors.append(ec)

                    strght_edges = np.array(new_edges, dtype=int)

                    loop_colors = [ec for ec, n in zip(loop_colors, self_loops)
                                   if n in restrict_targets]
                    self_loops  = self_loops.intersection(restrict_targets)

            if fast:
                if len(strght_edges):
                    dl = 0.5*np.max(nsize) if not simple_nodes else 0.

                    arrow_x  = pos[strght_edges[:, 1], 0] - \
                               pos[strght_edges[:, 0], 0]
                    arrow_x -= np.sign(arrow_x) * dl
                    arrow_y  = pos[strght_edges[:, 1], 1] - \
                               pos[strght_edges[:, 0], 1]
                    arrow_x -= np.sign(arrow_y) * dl

                    axis.quiver(
                        pos[strght_edges[:, 0], 0], pos[strght_edges[:, 0], 1],
                        arrow_x, arrow_y, scale_units='xy', angles='xy',
                        scale=1, alpha=ealpha, width=1.5e-3,
                        linewidths=0.5*esize, ec=ecolor, fc=ecolor, zorder=1)

                for i, s in enumerate(self_loops):
                    es = loop_sizes[i]
                    ec = loop_colors[i]
                    cs = _connectionstyle(axis, nsize[s], loop_sizes[i])

                    arrow = FancyArrowPatch(
                        pos[s], pos[s], arrowstyle=arrowstyle,
                        shrinkA=nsize[s], color=ec, linewidth=es,
                        connectionstyle=cs, zorder=1)

                    axis.add_patch(arrow)
            else:
                if len(strght_edges):
                    for i, (s, t) in enumerate(strght_edges):
                        xs, ys = pos[s, 0], pos[s, 1]
                        xt, yt = pos[t, 0], pos[t, 1]

                        if curved_edges:
                            arrow = FancyArrowPatch(
                                posA=(xs, ys), posB=(xt, yt),
                                arrowstyle=arrowstyle,
                                connectionstyle='arc3,rad=0.1',
                                alpha=ealpha, fc=ecolor[i], lw=0.5)
                            axis.add_patch(arrow)
                        else:
                            dl     = 0.5*nsize[t]
                            dx     = xt-xs
                            dx -= np.sign(dx) * dl
                            dy     = yt-ys
                            dy -= np.sign(dy) * dl
                            arrows.append(FancyArrow(
                                xs, ys, dx, dy, width=0.3*avg_size,
                                head_length=0.7*avg_size, head_width=0.7*avg_size,
                                length_includes_head=True, alpha=ealpha,
                                fc=strght_colors[i], lw=0.5))

                for i, s in enumerate(self_loops):
                    es = loop_sizes[i]
                    ec = loop_colors[i]

                    arrow = FancyArrowPatch(
                        pos[s], pos[s], arrowstyle=arrowstyle,
                        shrinkA=nsize[s], color=ec, linewidth=es,
                        connectionstyle='arc3,rad=1', zorder=1)

                    arrows.append(arrow)

        if not fast:
            arrows = PatchCollection(arrows, match_original=True, alpha=ealpha)
            arrows.set_zorder(1)
            axis.add_collection(arrows)

    if kwargs.get('tight', True):
        plt.tight_layout()
        plt.subplots_adjust(
            hspace=0., wspace=0., left=0., right=0.95 if colorbar else 1.,
            top=1., bottom=0.)

    if show:
        plt.show()


def hive_plot(network, radial, axes=None, axes_bins=None, axes_range=None,
              axes_angles=None, axes_labels=None, axes_units=None,
              intra_connections=True, highlight_nodes=None,
              highlight_edges=None, nsize=None, esize=None, max_nsize=10,
              max_esize=1, axes_colors=None, edge_colors=None, edge_alpha=0.05,
              nborder_color="k", nborder_width=0.2, show_names=True,
              show_circles=False, axis=None, tight=True, show=False):
    '''
    Draw a hive plot of the graph.

    Note
    ----
    For directed networks, the direction of intra-axis connections is
    counter-clockwise.
    For inter-axes connections, the default edge color is closest to the color
    of the source group (i.e. from a red group to a blue group, edge color will
    be a reddish violet , while from blue to red, it will be a blueish violet).

    Parameters
    ----------
    network : :class:`~nngt.Graph`
        Graph to plot.
    radial : str, list of str or array-like
        Values that will be used to place the nodes on the axes. Either one
        identical property is used for all axes (traditional hive plot) or
        one radial coordinate per axis is used (custom hive plot).
        If radial is a string or a list of strings, then these must correspond
        to the names of node attributes stored in the graph.
    axes : str, or list of str, optional (default: one per radial coordinate)
        Name of the attribute(s) that will be used to make each of the axes
        (i.e. each group of nodes).
        This can be either "groups" if the graph has a structure or is a
        :class:`~nngt.Network`, a list of (Meta)Group names, or any (list of)
        node attribute(s).
        If a single node attribute is used, `axes_bins` must be provided to
        make one axis for each range of values.
        If there are multiple radial coordinates, then leaving `axes` blanck
        will plot all nodes on each of the axes (one per radial coordinate).
    axes_bins : int or array-like, optional (default: all nodes on each axis)
        Required if there is a single radial coordinate and a single axis
        entry: provides the bins that will be used to separate the nodes
        into groups (one per axis). For N axes, there must therefore be N + 1
        entries in `axes_bins`, or `axis_bins` must be equal to N, in which
        case the nodes are separated into N evenly sized bins.
    axes_units : str, optional
        Units used to scale the axes. Either "native" to have them scaled
        between the minimal and maximal radial coordinates among all axes,
        "rank", to use the min and max ranks of the nodes on all axes, or
        "normed", to have each axis go from zero (minimal local radial
        coordinate) to one (maximal local radial coordinate).
        "native" is the default if there is a single radial coordinate,
        "normed" is the default for multiple coordinates.
    axes_angles : list of angles, optional (default: automatic)
        Angles for each of the axes, by increasing degree. If
        `intra_connections` is True, then angles of duplicate axes must be
        adjacent, e.g. ``[a1, a1bis, a2, a2bis, a3, a3bis]``.
    axes_labels : str or list of str, optional
        Label of each axis. For binned axes, it can be automatically formatted
        via the three entries ``{name}``, ``{start}``, ``{stop}``.
        E.g. "{name} in [{start}, {stop}]" would give "CC in [0, 0.2]" for
        a first axis and "CC in [0.2, 0.4]" for a second axis.
    intra_connections : bool, optional (default: True)
        Show connections between nodes belonging to the same axis. If true,
        then each axis is duplicated to display intra-axis connections.
    highlight_nodes : list of nodes, optional (default: all nodes)
        Highlight a subset of nodes and their connections, all other nodes
        and connections will be gray.
    highlight_edges : list of edges, optional (default: all edges)
        Highlight a subset of edges; all other connections will be gray.
    nsize : float, str, or array-like, optional (default: automatic)
        Size of the nodes on the axes. Either a fixed size, the name of a
        node attribute, or a list of user-defined values.
    esize : float or str, optional (default: 1)
        Size of the edges. Either a fixed size or the name of an edge
        attribute.
    max_nsize : float, optional (default: 10)
        Maximum node size if `nsize` is an attribute or a list of
        user-defined values.
    max_esize : float, optional (default: 1)
        Maximum edge size if `esize` is an attribute.
    axes_colors : valid matplotlib color/colormap, optional (default: Set1)
        Color associated to each axis.
    nborder_color : matplotlib color, optional (default: "k")
        Color of the node's border.
        or floats in [0, 1] defining the position in the palette.
    nborder_width : float, optional (default: 0.2)
        Width of the border.
    edge_colors : valid matplotlib color/colormap, optional (default: auto)
        Color of the edges. By default it is the intermediate color between
        two axes colors. To provide custom colors, they must be provided as
        a dictionnary of axes edges ``{(0, 0): "r", (0, 1): "g", (1, 0): "b"}``
        with default color being black.
    edge_alpha : float, optional (default: 0.05)
        Edge opacity.
    show_names : bool, optional (default: True)
        Show axes names and properties.
    show_circles : bool, optional (default: False)
        Show the circles associated to the maximum value of each axis.
    axis : matplotlib axis, optional (default: create new axis)
        Axis on which the network will be plotted.
    tight : bool, optional (default: True)
        Set figure layout to tight (set to False if plotting multiple axes on
        a single figure).
    show : bool, optional (default: True)
        Display the plot immediately.
    '''
    import matplotlib.pyplot as plt

    # get numer of axes and radial coordinates
    num_axes, num_radial = _get_axes_radial_coord(
        radial, axes, axes_bins, network)

    # get axes names, associated nodes, and radial values
    ax_names, ax_nodes, ax_radco = _get_axes_nodes(
        network, radial, axes, axes_bins, num_axes, num_radial)

    # get highlighted nodes and edges
    if highlight_nodes:
        highlight_nodes = set(highlight_nodes)
    else:
        highlight_nodes= set()

    if highlight_edges is not None:
        highlight_edges = {tuple(e) for e in highlight_edges}

    # get units, maximum values for the axes, renormalize radial values
    if axes_units is None:
        axes_units = "normed" if num_radial > 1 else "native"

    radial_values = _get_radial_values(ax_radco, axes_units, network)

    # compute the angles
    angles = None

    if axes_angles is None:
        dtheta = 2 * np.pi / num_axes

        if intra_connections:
            angles = []

            for i in range(num_axes):
                angles.extend(((i - 0.125)*dtheta, (i + 0.125)*dtheta))
        else:
            angles = [i*dtheta for i in range(num_axes)]
    else:
        angles = [a*np.pi/180 for a in ax_angles]

    # renormalize the sizes
    nsize = _get_size(nsize, max_nsize, ax_nodes, network)

    nedges = network.edge_nb()

    esize = np.ones(nedges) if esize is None else network.edge_attributes[esize]
    esize *= max_esize / esize.max()

    esize = {tuple(e): s for e, s in zip(network.edges_array, esize)}

    # get the colors
    ncolors, ecolors = _get_colors(axes_colors, edge_colors, angles, num_axes,
                                   intra_connections, network)

    # make the figure
    if axis is None:
        _, axis = plt.subplots()

    # plot the nodes and axes
    node_pos  = []
    max_radii = []

    for i, (nn, rr) in enumerate(zip(ax_nodes, radial_values)):
        if len(nn):
            # max radii
            rax = np.array([RMIN, rr[nn].max()])

            max_radii.extend([rax[-1]]*(1 + intra_connections))

            # plot max radii
            if show_circles:
                aa = np.arange(0, 2*np.pi, 0.02)
                xx = rax[-1]*np.cos(aa)
                yy = rax[-1]*np.sin(aa)
                axis.plot(xx, yy, color="grey", alpha=0.2, zorder=1)

            # comppute angles
            aa = [angles[2*i] if intra_connections else angles[i]]

            if intra_connections:
                aa += [angles[2*i+1]]

            for j, a in enumerate(aa):
                # plot axes lines
                lw = 1 if j % 2 else 2

                axis.plot(rax*np.cos(a), rax*np.sin(a), color="grey", lw=lw,
                          zorder=1)

                # compute node positions
                xx = rr*np.cos(a)
                yy = rr*np.sin(a)

                node_pos.append(np.array([xx, yy]).T)

                if highlight_nodes:
                    greys = list(set(nn).difference(highlight_nodes))

                    _plot_nodes(greys, nsize, xx, yy, "grey",
                                nborder_width, nborder_color, axis, zorder=3)

                hlght = (nn if not highlight_nodes
                         else list(highlight_nodes.intersection(nn)))

                _plot_nodes(hlght, nsize, xx, yy, ncolors[i],
                            nborder_width, nborder_color, axis, zorder=4)
        else:
            node_pos.extend([[]]*(1 + intra_connections))
            max_radii.extend([RMIN]*(1 + intra_connections))

    # plot the edges
    xs, ys = [], []

    for i, n1 in enumerate(ax_nodes):
        targets = ax_nodes if network.is_directed() else ax_nodes[i:]

        for j, n2 in enumerate(ax_nodes):
            # ignore i = j if intra_connections is True
            if i == j and not intra_connections:
                continue

            # find which axes should be used
            idx_s, idx_t = _get_ax_angles(
                angles, i, j, intra_connections)

            # get the edges
            edges = network.get_edges(source_node=n1, target_node=n2)

            if len(edges):
                color = ecolors[(i, j)]

                paths_greys = []
                paths_hghlt = []

                lw = []

                for (ns, nt) in edges:
                    pstart = node_pos[idx_s][ns]
                    pstop  = node_pos[idx_t][nt]

                    contains = True

                    if highlight_edges is not None:
                        contains = (ns, nt) in highlight_edges
                    elif highlight_nodes is not None:
                        contains = \
                            ns in highlight_nodes or nt in highlight_nodes

                    if highlight_edges is None or contains:
                        paths_hghlt.append(_plot_bezier(
                            pstart, pstop, angles[idx_s], angles[idx_t],
                            radial_values[i][ns], radial_values[j][nt], i, j,
                            num_axes, xs, ys))

                        lw.append(esize[(ns, nt)])
                    else:
                        paths_greys.append(_plot_bezier(
                            pstart, pstop, angles[idx_s], angles[idx_t],
                            radial_values[i][ns], radial_values[j][nt], i, j,
                            num_axes, xs, ys))

                if paths_greys:
                    pcol = PathCollection(
                        paths_greys, facecolors="none", edgecolors="grey",
                        alpha=0.1*edge_alpha, zorder=1)

                    axis.add_collection(pcol)

                alpha = 0.7 if highlight_nodes else edge_alpha

                pcol = PathCollection(paths_hghlt, facecolors="none", lw=lw,
                                      edgecolors=color, alpha=alpha, zorder=2)

                axis.add_collection(pcol)

    _set_names_lims(ax_names, angles, max_radii, xs, ys, intra_connections,
                    show_names, axis, show_circles)


    axis.set_aspect(1)
    axis.axis('off')

    if tight:
        plt.tight_layout()

    if show:
        plt.show()


def library_draw(network, nsize="total-degree", ncolor="group", nshape="o",
                 nborder_color="k", nborder_width=0.5, esize=1., ecolor="k",
                 ealpha=0.5, max_nsize=5., max_esize=2., curved_edges=False,
                 threshold=0.5, decimate_connections=None, spatial=True,
                 restrict_sources=None, restrict_targets=None,
                 restrict_nodes=None, restrict_edges=None,
                 show_environment=True, size=(600, 600), xlims=None,
                 ylims=None, dpi=75, axis=None, colorbar=False,
                 show_labels=False, layout=None, show=False, **kwargs):
    '''
    Draw a given :class:`~nngt.Graph` using the underlying library's drawing
    functions.

    .. versionadded:: 2.0

    .. warning::
        When using igraph or graph-tool, if you want to use the `axis`
        argument, then you must first switch the matplotlib backend to its
        cairo version using e.g. ``plt.switch_backend("Qt5Cairo")`` if your
        normal backend is Qt5 ("Qt5Agg").

    Parameters
    ----------
    network : :class:`~nngt.Graph` or subclass
        The graph/network to plot.
    nsize : float, array of float or string, optional (default: "total-degree")
        Size of the nodes as a percentage of the canvas length. Otherwise, it
        can be a string that correlates the size to a node attribute among
        "in/out/total-degree", or "betweenness".
    ncolor : float, array of floats or string, optional (default: 0.5)
        Color of the nodes; if a float in [0, 1], position of the color in the
        current palette, otherwise a string that correlates the color to a node
        attribute among "in/out/total-degree", "betweenness" or "group".
    nshape : char, array of chars, or groups, optional (default: "o")
        Shape of the nodes (see `Matplotlib markers <http://matplotlib.org/api/
        markers_api.html?highlight=marker#module-matplotlib.markers>`_).
        When using groups, they must be pairwise disjoint; markers will be
        selected iteratively from the matplotlib default markers.
    nborder_color : char, float or array, optional (default: "k")
        Color of the node's border using predefined `Matplotlib colors
        <http://matplotlib.org/api/colors_api.html?highlight=color
        #module-matplotlib.colors>`_).
        or floats in [0, 1] defining the position in the palette.
    nborder_width : float or array of floats, optional (default: 0.5)
        Width of the border in percent of canvas size.
    esize : float, str, or array of floats, optional (default: 0.5)
        Width of the edges in percent of canvas length. Available string values
        are "betweenness" and "weight".
    ecolor : str, char, float or array, optional (default: "k")
        Edge color. If ecolor="groups", edges color will depend on the source
        and target groups, i.e. only edges from and toward same groups will
        have the same color.
    max_esize : float, optional (default: 5.)
        If a custom property is entered as `esize`, this normalizes the edge
        width between 0. and `max_esize`.
    threshold : float, optional (default: 0.5)
        Size under which edges are not plotted.
    decimate_connections : int, optional (default: keep all connections)
        Plot only one connection every `decimate_connections`.
        Use -1 to hide all edges.
    spatial : bool, optional (default: True)
        If True, use the neurons' positions to draw them.
    restrict_sources : str, group, or list, optional (default: all)
        Only draw edges starting from a restricted set of source nodes.
    restrict_targets : str, group, or list, optional (default: all)
        Only draw edges ending on a restricted set of target nodes.
    restrict_nodes : str, group, or list, optional (default: plot all nodes)
        Only draw a subset of nodes.
    restrict_edges : list of edges, optional (default: all)
        Only draw a subset of edges.
    show_environment : bool, optional (default: True)
        Plot the environment if the graph is spatial.
    fast : bool, optional (default: False)
        Use a faster algorithm to plot the edges. This method leads to less
        pretty plots and zooming on the graph will make the edges start or
        ending in places that will differ more or less strongly from the actual
        node positions.
    size : tuple of ints, optional (default: (600, 600))
        (width, height) tuple for the canvas size (in px).
    dpi : int, optional (default: 75)
        Resolution (dot per inch).
    colorbar : bool, optional (default: False)
        Whether to display a colorbar for the node colors or not.
    axis : matplotlib axis, optional (default: create new axis)
        Axis on which the network will be plotted.
    layout : str, optional (default: library-dependent or spatial positions)
        Name of a standard layout to structure the network. Available layouts
        are: "circular", "spring-block", "random". If no layout is
        provided and the network is spatial, then node positions will be
        used by default.
    show : bool, optional (default: True)
        Display the plot immediately.
    **kwargs : dict
        Optional keyword arguments including `node_cmap` to set the
        nodes colormap (default is "magma" for continuous variables and
        "Set1" for groups) and the boolean `simple_nodes` to make node
        plotting faster.
    '''
    import matplotlib.pyplot as plt

    # backend and axis
    if nngt.get_config("backend") in ("graph-tool", "igraph"):
        mpl_backend = mpl.get_backend()

        if mpl_backend.startswith("Qt4"):
            if mpl_backend != "Qt4Cairo":
                plt.switch_backend("Qt4Cairo")
        elif mpl_backend.startswith("Qt5"):
            if mpl_backend != "Qt5Cairo":
                plt.switch_backend("Qt5Cairo")
        elif mpl_backend.startswith("GTK"):
            if mpl_backend != "GTK3Cairo":
                plt.switch_backend("GTK3Cairo")
        elif mpl_backend != "cairo":
            plt.switch_backend("cairo")

    if axis is None:
        size_inches = (size[0]/float(dpi), size[1]/float(dpi))
        fig, axis = plt.subplots(figsize=size_inches)

    axis.axis('off')

    # default plot
    if nngt.get_config("backend") == "nngt":
        draw_network(
            network, nsize=nsize, ncolor=ncolor, nshape=nshape,
            nborder_color=nborder_color, nborder_width=nborder_width,
            esize=esize, ecolor=ecolor, ealpha=ealpha, max_nsize=max_nsize,
            max_esize=max_esize, curved_edges=curved_edges,
            threshold=threshold, decimate_connections=decimate_connections,
            spatial=spatial, restrict_nodes=restrict_nodes,
            show_environment=show_environment, size=size, axis=axis,
            layout=layout, show=show, **kwargs)

        return

    # otherwise, preapre data
    restrict_nodes = _convert_to_nodes(restrict_nodes,
                                       "restrict_nodes", network)

    # shize and shape
    markers, nsize, esize = _node_edge_shape_size(
        network, nshape, nsize, max_nsize, esize, max_esize, restrict_nodes,
        restrict_edges, size, threshold)

    # node color information
    default_ncmap = (palette_discrete() if not nonstring_container(ncolor) and
                     ncolor == "group" else palette_continuous())

    ncmap = get_cmap(kwargs.get("node_cmap", default_ncmap))

    node_color, nticks, ntickslabels, nlabel = \
        _node_color(network, restrict_nodes, ncolor)

    # edge color
    ecolor = _edge_prop(network, ecolor)
    esize  = _edge_prop(network, esize)

    if nonstring_container(esize) and len(esize):
        esize *= max_esize / np.max(esize)
    
    # environment
    if spatial and network.is_spatial():
        if show_environment:
            nngt.geometry.plot.plot_shape(network.shape, axis=axis, show=False)

    # do the plot
    if nngt.get_config("backend") == "graph-tool":
        from graph_tool.draw import (graph_draw, sfdp_layout, random_layout)

        graph = network.graph

        # resize
        if nonstring_container(nsize):
            nsize *= 0.05

        nborder_width *= 0.1

        esize *= 0.02

        # positions
        pos = None

        if layout is None:
            if isinstance(network, nngt.SpatialGraph) and spatial:
                xy  = network.get_positions()
                pos = graph.new_vp("vector<double>", vals=xy)
            else:
                weights = (None if not network.is_weighted()
                           else graph.edge_properties['weight'])
                pos = sfdp_layout(graph, eweight=weights)
        elif layout == "random":
            pos = random_layout(graph)
        elif layout == "circular":
            pos = graph.new_vp("vector<double>",
                               vals=_circular_layout(network, nsize))
        elif nonstring_container(layout):
            assert np.shape(layout) == (network.node_nb(), 2), \
                "One position per node in the network is required."
            pos = graph.new_vp("vector<double>", vals=layout)
        else:
            # spring block
            weights = (None if not network.is_weighted()
                       else graph.edge_properties['weight'])
            pos = sfdp_layout(graph, eweight=weights)

        convert_shape = {
            "o": "circle",
            "v": "triangle",
            "^": "triangle",
            "s": "square",
            "p": "pentagon",
            "h": "hexagon",
            "H": "hexagon",
        }

        shape_dict = defaultdict(
            lambda k: "circle" if k not in convert_shape.values() else k)

        for k, v in convert_shape.items():
            shape_dict[k] = v

        vprops = {
            "shape": shape_dict[nshape],
            "fill_color": _to_gt_prop(graph, node_color, ncmap, color=True),
            "color": _to_gt_prop(graph, nborder_color, ncmap, color=True),
            "size": _to_gt_prop(graph, nsize, ncmap),
            "pen_width": _to_gt_prop(graph, nborder_width, ncmap),
        }

        if vprops["fill_color"] is None:
            vprops["fill_color"] = [0.640625, 0, 0, 0.9]

        eprops = None if network.edge_nb() == 0 else {
            "color": _to_gt_prop(graph, ecolor, palette_continuous(),
                                 ptype='edge', color=True),
            "pen_width": _to_gt_prop(graph, esize, None, ptype='edge'),
        }

        if restrict_edges is not None:
            efilt = network.graph.new_ep(
                "bool", vals=np.zeros(network.edge_nb(), dtype=bool))
            eids = [network.edge_id(e) for e in restrict_edges]

            efilt.a[eids] = 1

            network.graph.set_edge_filter(efilt)

        graph_draw(network.graph, pos=pos, vprops=vprops, eprops=eprops,
                   output_size=size, mplfig=axis)

        if restrict_edges is not None:
            # clear edge filter
            network.graph.set_edge_filter(None)
    elif nngt.get_config("backend") == "networkx":
        import networkx as nx

        pos = None

        if layout is None:
            if isinstance(network, nngt.SpatialGraph) and spatial:
                xy  = network.get_positions()
                pos = {i: coords for i, coords in enumerate(xy)}
        elif layout == "circular":
            pos = nx.circular_layout(network.graph)
        elif layout == "random":
            pos = nx.random_layout(network.graph)
        elif nonstring_container(layout):
            assert np.shape(layout) == (network.node_nb(), 2), \
                "One position per node in the network is required."
            pos = {i: coords for i, coords in enumerate(layout)}
        else:
            pos = nx.spring_layout(network.graph)

        # normalize sizes compared to igraph
        nsize = _increase_nx_size(nsize)

        nborder_width = _increase_nx_size(nborder_width, 2)

        edges = None if restrict_edges is None else list(restrict_edges)

        nx.draw_networkx(
            network.graph, pos=pos, ax=axis, nodelist=restrict_nodes,
            edgelist=edges, node_size=nsize, node_color=node_color,
            node_shape=nshape, linewidths=nborder_width, edge_color=ecolor,
            edge_cmap=palette_continuous(), cmap=ncmap,
            with_labels=show_labels, width=esize, edgecolors=nborder_color)
    elif nngt.get_config("backend") == "igraph":
        from igraph import Layout, PrecalculatedPalette

        pos = None

        if layout is None:
            if isinstance(network, nngt.SpatialGraph) and spatial:
                xy  = network.get_positions()
                pos = Layout(xy)
        elif layout == "circular":
            pos = network.graph.layout_circle()
        elif layout == "random":
            pos = network.graph.layout_random()

        palette = PrecalculatedPalette(ncmap(np.linspace(0, 1, 256)))

        # convert color to igraph-format
        node_color = _to_ig_color(node_color)
        ecolor     = _to_ig_color(ecolor)

        convert_shape = {
            "o": "circle",
            "v": "triangle-down",
            "^": "triangle-up",
            "s": "rectangle",
        }

        shape_dict = defaultdict(
            lambda k: "circle" if k not in convert_shape.values() else k)

        for k, v in convert_shape.items():
            shape_dict[k] = v

        if nonstring_container(nsize):
            nsize = list(nsize)

        if nonstring_container(node_color):
            node_color = list(node_color)

        if nonstring_container(esize):
            esize = list(esize)

        if nonstring_container(ecolor):
            ecolor = list(ecolor)

        visual_style = {
            "vertex_size": nsize,
            "vertex_color": node_color,
            "vertex_shape": shape_dict[nshape],
            "edge_width": esize,
            "edge_color": ecolor,
            "layout": pos,
            "palette": palette,
        }

        graph = network.graph

        if restrict_edges is not None:
            eids  = [network.edge_id(e) for e in restrict_edges]
            graph = network.graph.subgraph_edges(eids, delete_vertices=False)

        graph_artist = GraphArtist(graph, axis, **visual_style)

        axis.artists.append(graph_artist)

    if "title" in kwargs:
        axis.set_title(kwargs["title"])

    if show:
        plt.show()


def chord_diagram(network, weights=True, names=None, order=None, width=0.1,
                  pad=2., gap=0.03, chordwidth=0.7, axis=None, colors=None,
                  cmap=None, alpha=0.7, use_gradient=False, chord_colors=None,
                  show=False, **kwargs):
    """
    Plot a chord diagram.

    Parameters
    ----------
    network : a :class:`nngt.Graph` object
        Network used to plot the chord diagram.
    weights : bool or str, optional (default: 'weight' attribute)
        Weights used to plot the connections.
    names : str or list of str, optional (default: no names)
        Names of the nodes that will be displayed, either a node attribute
        or a custom list (must be ordered following the nodes' indices).
    order : list, optional (default: order of the matrix entries)
        Order in which the arcs should be placed around the trigonometric
        circle.
    width : float, optional (default: 0.1)
        Width/thickness of the ideogram arc.
    pad : float, optional (default: 2)
        Distance between two neighboring ideogram arcs. Unit: degree.
    gap : float, optional (default: 0.03)
        Distance between the arc and the beginning of the cord.
    chordwidth : float, optional (default: 0.7)
        Position of the control points for the chords, controlling their shape.
    axis : matplotlib axis, optional (default: new axis)
        Matplotlib axis where the plot should be drawn.
    colors : list, optional (default: from `cmap`)
        List of user defined colors or floats.
    cmap : str or colormap object (default: viridis)
        Colormap that will be used to color the arcs and chords by default.
        See `chord_colors` to use different colors for chords.
    alpha : float in [0, 1], optional (default: 0.7)
        Opacity of the chord diagram.
    use_gradient : bool, optional (default: False)
        Whether a gradient should be use so that chord extremities have the
        same color as the arc they belong to.
    chord_colors : str, or list of colors, optional (default: None)
        Specify color(s) to fill the chords differently from the arcs.
        When the keyword is not used, chord colors default to the colomap given
        by `colors`.
        Possible values for `chord_colors` are:

        * a single color (do not use an RGB tuple, use hex format instead),
          e.g. "red" or "#ff0000"; all chords will have this color
        * a list of colors, e.g. ``["red", "green", "blue"]``, one per node
          (in this case, RGB tuples are accepted as entries to the list).
          Each chord will get its color from its associated source node, or
          from both nodes if `use_gradient` is True.
    show : bool, optional (default: False)
        Whether the plot should be displayed immediately via an automatic call
        to `plt.show()`.
    kwargs : keyword arguments
        Available kwargs are:

        ================  ==================  ===============================
              Name               Type           Purpose and possible values
        ================  ==================  ===============================
        fontcolor         str or list         Color of the names
        fontsize          int                 Size of the font for names
        rotate_names      (list of) bool(s)   Rotate names by 90°
        sort              str                 Either "size" or "distance"
        zero_entry_size   float               Size of zero-weight reciprocal
        ================  ==================  ===============================
    """
    ww = 'weight' if weights is True else weights
    nn = network.node_attributes[names] if isinstance(names, str) else names

    mat = network.adjacency_matrix(weights=ww)

    return _chord_diag(
        mat, nn, order=order, width=width, pad=pad, gap=gap,
        chordwidth=chordwidth, ax=axis, colors=colors, cmap=cmap, alpha=alpha,
        use_gradient=use_gradient, chord_colors=chord_colors, show=show,
        **kwargs)


# ----- #
# Tools #
# ----- #

def _node_edge_shape_size(network, nshape, nsize, max_nsize, esize, max_esize,
                          restrict_nodes, edges, size, threshold,
                          simple_nodes=False):
    ''' Returns the shape and size of the nodes and edges '''
    n = network.node_nb() if restrict_nodes is None else len(restrict_nodes)
    e = len(edges) if edges is not None else network.edge_nb()

    # markers
    markers = nshape

    if nonstring_container(nshape):
        if isinstance(nshape[0], nngt.Group):
            # check disjunction
            for i, g in enumerate(nshape):
                for j in range(i + 1, len(nshape)):
                    if not set(g.ids).isdisjoint(nshape[j].ids):
                        raise ValueError("Groups passed to `nshape` "
                                         "must be disjoint.")

            mm = cycle(MarkerStyle.filled_markers)

            shapes  = np.full(network.node_nb(), "", dtype=object)

            for g, m in zip(nshape, mm):
                shapes[g.ids] = m

            markers = list(shapes)
        elif len(nshape) != network.node_nb():
            raise ValueError("When passing an array of markers to "
                             "`nshape`, one entry per node in the "
                             "network must be provided.")
    else:
        markers = [nshape for _ in range(network.node_nb())]

    # size
    if isinstance(nsize, str):
        if e:
            nsize = _node_size(network, restrict_nodes, nsize)
            nsize *= max_nsize / np.max(nsize)
        else:
            nsize = np.ones(n, dtype=float)
    elif isinstance(nsize, (float, int, np.number)):
        nsize = np.full(n, nsize, dtype=float)
    elif nonstring_container(nsize):
        if len(nsize) == n:
            nsize *= max_nsize / np.max(nsize)
        elif len(nsize) == network.node_nb() and restrict_nodes is not None:
            nsize = np.asarray(nsize)[list(restrict_nodes)]
            nsize /= max_nsize / np.max(nsize)
        else:
            raise ValueError("`nsize` must contain either one entry per node "
                             "or be the same length as `restrict_nodes`.")

    nsize *= 0.01 * size[0]

    if e:
        if isinstance(esize, str):
            esize  = _edge_size(network, edges, esize)
            esize *= max_esize
            esize[esize < threshold] = 0.

        esize *= 0.005 * size[0]  # border on each side (so 0.5 %)
    else:
        esize = np.array([])

    return markers, nsize, esize


def _set_ax_lim(ax, xdata, ydata, xlims, ylims):
    if xlims is not None:
        ax.set_xlim(*xlims)
    else:
        x_min, x_max = np.min(xdata), np.max(xdata)
        width = x_max - x_min
        ax.set_xlim(x_min - 0.05*width, x_max + 0.05*width)
    if ylims is not None:
        ax.set_ylim(*ylims)
    else:
        y_min, y_max = np.min(ydata), np.max(ydata)
        height = y_max - y_min
        ax.set_ylim(y_min - 0.05*height, y_max + 0.05*height)


def _node_size(network, restrict_nodes, nsize):
    restrict_nodes = None if restrict_nodes is None else list(restrict_nodes)

    n = network.node_nb() if restrict_nodes is None else len(restrict_nodes)

    size = np.ones(n, dtype=float)

    if "degree" in nsize:
        deg_type = nsize[:nsize.index("-")]
        size = network.get_degrees(deg_type,
                                   nodes=restrict_nodes).astype(float)
        if np.isclose(size.min(), 0):
            size[np.isclose(size, 0)] = 0.5
        if size.max() > 15*size.min():
            size = np.power(size, 0.4)
    elif "strength" in nsize:
        deg_type = nsize[:nsize.index("-")]
        size = network.get_degrees(deg_type, weights='weight',
                                   nodes=restrict_nodes)
        if np.isclose(size.min(), 0):
            size[np.isclose(size, 0)] = 0.5
        if size.max() > 15*size.min():
            size = np.power(size, 0.4)
    elif nsize == "betweenness":
        betw = None

        if restrict_nodes is None:
            betw = network.get_betweenness("node").astype(float)
        else:
            betw = network.get_betweenness(
                "node").astype(float)[restrict_nodes]

        if network.is_connected("weak") == 1:
            size *= betw
            if size.max() > 15*size.min():
                min_size = size[size!=0].min()
                size[size == 0.] = min_size
                size = np.log(size)
                if size.min()<0:
                    size -= 1.1*size.min()
    elif nsize == "clustering":
        size *= nngt.analysis.local_clustering(network, nodes=restrict_nodes)
    elif nsize in nngt.analyze_graph:
        if restrict_nodes is None:
            size *= nngt.analyze_graph[nsize](network)
        else:
            size *= nngt.analyze_graph[nsize](network)[restrict_nodes]

    if np.any(size):
        size /= size.max()

    return size.astype(float)


def _edge_size(network, edges, esize):
    num_edges = len(edges) if edges is not None else network.edge_nb()

    size = np.repeat(1., num_edges)

    if num_edges:
        max_size = 1.

        if nonstring_container(esize):
            max_size = np.max(esize)
        elif esize == "betweenness":
            betw = network.get_betweenness("edge")

            max_size = np.max(betw)

            size = betw if restrict_nodes is None else betw[restrict_nodes]
        elif esize == "weight":
            size = network.get_weights(edges=edges)

            max_size = np.max(network.get_weights())

        if np.any(size):
            size /= max_size

    return size


def _node_color(network, restrict_nodes, ncolor):
    '''
    Return an array of colors, a set of ticks, and a label for the colorbar
    of the nodes (if necessary).
    '''
    color        = ncolor
    nticks       = None
    ntickslabels = None
    nlabel       = ""

    n = network.node_nb() if restrict_nodes is None else len(restrict_nodes)

    if restrict_nodes is not None:
        restrict_nodes = set(restrict_nodes)

    if isinstance(ncolor, float):
        color = np.repeat(ncolor, n)
    elif isinstance(ncolor, str):
        if ncolor == "group" or ncolor == "groups":
            color = np.zeros(n)
            if network.structure is not None:
                l = len(network.structure)
                c = np.linspace(0, 1, l)
                tmp = 0
                for i, group in enumerate(network.structure.values()):
                    if restrict_nodes is None:
                        color[group.ids] = c[i]
                    else:
                        ids = restrict_nodes.intersection(group.ids)
                        for j in range(len(ids)):
                            color[tmp + j] = c[i]
                        tmp += len(ids)

                nlabel       = "Neuron groups"
                nticks       = list(range(len(network.structure)))
                ntickslabels = [s.replace("_", " ")
                                for s in network.structure.keys()]
        else:
            values = None
            if "degree" in ncolor:
                dtype   = ncolor[:ncolor.find("-")]
                values = network.get_degrees(dtype, nodes=restrict_nodes)
            elif ncolor == "betweenness":
                if restrict_nodes is None:
                    values = network.get_betweenness("node")
                else:
                    values = network.get_betweenness(
                        "node")[list(restrict_nodes)]
            elif ncolor in network.node_attributes:
                values = network.get_node_attributes(
                    name=ncolor, nodes=restrict_nodes)
            elif ncolor == "clustering" :
                values = nngt.analysis.local_clustering(
                    network, nodes=restrict_nodes)
            elif ncolor in nngt.analyze_graph:
                if restrict_nodes is None:
                    values = nngt.analyze_graph[ncolor](network)
                else:
                    values = nngt.analyze_graph[ncolor](
                        network)[list(restrict_nodes)]
            elif ncolor in ColorConverter.colors or ncolor.startswith("#"):
                color = np.repeat(ncolor, n)
            else:
                raise RuntimeError("Invalid `ncolor`: {}.".format(ncolor))

            if values is not None:
                vmin, vmax = np.min(values), np.max(values)
                #~ color = (values - vmin) / (vmax - vmin)
                color = values

                nlabel = "Node " + ncolor.replace("_", " ")
                setval = set(values)
                if len(setval) <= 10:
                    nticks = list(setval)
                    nticks.sort()
                    ntickslabels = nticks
                else:
                    nticks       = np.linspace(vmin, vmax, 10)
                    ntickslabels = nticks
    else:
        nlabel  = "Custom node colors"
        uniques = np.unique(ncolor, axis=0)
        if len(uniques) <= 10:
            nticks = uniques
        else:
            nticks = np.linspace(np.min(ncolor), np.max(ncolor), 10)
        ntickslabels = nticks

    return color, nticks, ntickslabels, nlabel


def _edge_prop(network, value):
    prop = value

    enum = network.edge_nb()

    if isinstance(value, str) and value not in ColorConverter.colors:
        if value in network.edge_attributes:
            color = network.edge_attributes[value]
        elif value == "betweenness":
            prop = network.get_betweenness("edge")
        else:
            raise RuntimeError("Invalid `value`: {}.".format(value))

    return prop


def _discrete_cmap(N, base_cmap=None, clist=None):
    '''
    Create an N-bin discrete colormap from the specified input map

    Parameters
    ----------
    N : number of values
    base_cmap : str, None, or cmap object
    clist : list of colors

    # Modified from Jake VanderPlas
    # License: BSD-style
    '''
    import matplotlib.pyplot as plt
    # Note that if base_cmap is a string or None, you can simply do
    #    return plt.cm.get_cmap(base_cmap, N)
    # The following works for string, None, or a colormap instance:
    base = plt.cm.get_cmap(base_cmap, N)
    color_list = base(np.linspace(0, 1, N)) if clist is None else clist
    cmap_name = base.name + str(N)
    try:
        return base.from_list(cmap_name, color_list, N)
    except:
        return ListedColormap(color_list, cmap_name, N=N)


def _convert_to_nodes(node_restriction, name, network):
    if nonstring_container(node_restriction):
        if isinstance(node_restriction[0], str):
            assert network.structure is not None, \
                "`" + name + "` can be string only for Network or graph " \
                "with a `structure`."
            ids = set()
            for name in node_restriction:
                ids.update(network.structure[name].ids)
            return ids
        elif isinstance(node_restriction[0], nngt.Group):
            ids = set()
            for g in node_restriction:
                ids.update(g.ids)
            return ids

        return set(node_restriction) 
    elif isinstance(node_restriction, str):
        assert network.is_network(), \
            "`" + name + "` can be string only for Network."
        return set(network.structure[node_restriction].ids)
    elif isinstance(node_restriction, nngt.Group):
        return set(node_restriction.ids)
    elif node_restriction is not None:
        raise ValueError(
            "Invalid `" + name + "`: '{}'".format(node_restriction))

    return node_restriction


def _custom_arrows(sources, targets, angle):
    '''
    Create a curved arrow between `source` and `target` as the combination of
    the arc of a circle and a triangle.

    The initial and final angle $\alpha$ between the source-target line and
    the arrow is linked to the radius of the circle, $r$ and the distance $d$
    between the points:

    .. math:: r = \frac{d}{2 \cdot \tan(\alpha)}

    The beginning and the end of the arc are given through initial and final
    angles, respectively $\theta_1$ and $\theta_2$, which are given with
    respect to the y-axis; This leads to $\alpha = 0.5(\theta_1 - \theta_2)$.
    '''
    # compute the distances between the points
    pass
    #~ # compute the radius and the position of the center of the circle

    #~ #========Line
    #~ arc = Arc([centX,centY],radius,radius,angle=angle_,
          #~ theta1=0,theta2=theta2_,capstyle='round',linestyle='-',lw=10,color=color_)
    #~ ax.add_patch(arc)


    #~ #========Create the arrow head
    #~ endX=centX+(radius/2)*np.cos(rad(theta2_+angle_)) #Do trig to determine end position
    #~ endY=centY+(radius/2)*np.sin(rad(theta2_+angle_))

    #~ ax.add_patch(                    #Create triangle as arrow head
        #~ RegularPolygon(
            #~ (endX, endY),            # (x,y)
            #~ 3,                       # number of vertices
            #~ radius/9,                # radius
            #~ rad(angle_+theta2_),     # orientation
            #~ color=color_
        #~ )
    #~ )


def _to_ig_color(color):
    import igraph as ig

    if isinstance(color, str) and color not in ig.known_colors:
        color = str(ColorConverter.to_rgb(color))[1:-1]
    elif nonstring_container(color) and len(color):
        # need to convert floating point colors to [0, 255] integers
        if is_integer(color[0]) or isinstance(color[0], float):
            vmin = np.min(color)
            vmax = np.max(color)
            vint = vmax - vmin
            if vint > 0:
                color = [int(255 * (v - vmin) / vint) for v in color]
            else:
                color = [0]*len(color)
        else:
            for i, c in enumerate(color):
                if isinstance(color, str) and color not in ig.known_colors:
                    color[i] = str(ColorConverter.to_rgb(color))[1:-1]

    return color


def _increase_nx_size(size, factor=4):
    
    if isinstance(size, float) or is_integer(size):
        return factor*size
    elif nonstring_container(size) and len(size):
        if isinstance(size[0], float) or is_integer(size[0]):
            return factor*np.asarray(size)

    return size


def _to_gt_prop(graph, value, cmap, ptype='node', color=False):
    pmap = (graph.new_vertex_property if ptype == 'node'
            else graph.new_edge_property)
    
    if nonstring_container(value) and len(value):
        if isinstance(value[0], str):
            if color:
                # custom namedcolors
                return pmap("vector<double>",
                            vals=[ColorConverter.to_rgba(v) for v in value])
            else:
                return pmap("string", vals=value)
        elif nonstring_container(value[0]):
            # direct rgb(a) description
            return pmap("vector<double>", vals=value)

        # numbers
        if color:
            vmin, vmax = np.min(value), np.max(value)

            normalized = None

            if vmax - vmin > 0:
                normalized = (np.array(value) - vmin) / (vmax - vmin)
            else:
                return normalized

            return pmap("vector<double>", vals=[cmap(v) for v in normalized])

        return pmap("double", vals=value)
        

    return value


class GraphArtist(Artist):
    """
    Matplotlib artist class that draws igraph graphs.

    Only Cairo-based backends are supported.

    Adapted from: https://stackoverflow.com/a/36154077/5962321
    """

    def __init__(self, graph, axis, palette=None, *args, **kwds):
        """Constructs a graph artist that draws the given graph within
        the given bounding box.

        `graph` must be an instance of `igraph.Graph`.
        `bbox` must either be an instance of `igraph.drawing.BoundingBox`
        or a 4-tuple (`left`, `top`, `width`, `height`). The tuple
        will be passed on to the constructor of `BoundingBox`.
        `palette` is an igraph palette that is used to transform
        numeric color IDs to RGB values. If `None`, a default grayscale
        palette is used from igraph.

        All the remaining positional and keyword arguments are passed
        on intact to `igraph.Graph.__plot__`.
        """
        from igraph import BoundingBox, palettes

        super().__init__()

        self.graph = graph
        self.palette = palette or palettes["gray"]
        self.bbox = BoundingBox(axis.bbox.bounds)
        self.args = args
        self.kwds = kwds

    def draw(self, renderer):
        from matplotlib.backends.backend_cairo import RendererCairo

        if not isinstance(renderer, RendererCairo):
            raise TypeError(
                "graph plotting is supported only on Cairo backends")

        self.graph.__plot__(renderer.gc.ctx, self.bbox, self.palette,
                            *self.args, **self.kwds)


def _circular_layout(graph, node_size):
    max_nsize = np.max(node_size)

    # chose radius such that r*dtheta > max_nsize
    dtheta = 2*np.pi / graph.node_nb()

    r = 1.1*max_nsize / dtheta

    thetas = np.array([i*dtheta for i in range(graph.node_nb())])
    x = r*np.cos(thetas)
    y = r*np.sin(thetas)

    return np.array((x, y)).T



def _connectionstyle(axis, nsize, esize):
    def cs(posA, posB, *args, **kwargs):
        # check if we need to do a self-loop
        if np.all(posA == posB):
            # Self-loops are scaled by node size
            vshift = 0.1*max(nsize, 2*esize)
            hshift = 0.7*vshift
            # this is called with _screen space_ values so covert back
            # to data space

            s1 = np.asarray([-hshift, vshift])
            s2 = np.asarray([hshift, vshift])

            p1 = axis.transData.inverted().transform(posA)
            p2 = axis.transData.inverted().transform(posA + s1)
            p3 = axis.transData.inverted().transform(posA + s2)

            path = [p1, p2, p3, p1]

            ret = mpl.path.Path(axis.transData.transform(path), [1, 2, 2, 2])
        # if not, fall back to the user specified behavior
        else:
            return None

        return ret

    return cs

def _split_edges_sizes(edges, esize, decimate_connections, ecolor=None,
                       strght_colors=None, loop_colors=None):
    strght_edges, self_loops = None, None
    strght_sizes, loop_sizes = None, None

    keep  = (esize > 0) if nonstring_container(esize) else True
    loops = (edges[:, 0] == edges[:, 1])

    strght = keep*(~loops)

    strght_edges = edges[strght]

    self_loops = set(edges[loops, 0])

    if ecolor is not None:
        if nonstring_container(ecolor):
            if decimate_connections < 1:
                strght_colors.extend(ecolor[strght])
            loop_colors.extend(ecolor[loops])
        else:
            if decimate_connections < 1:
                strght_colors.extend([ecolor]*len(strght_edges))
            loop_colors.extend([ecolor]*len(self_loops))

    if nonstring_container(esize):
        strght_sizes = esize[strght]
        loop_sizes = esize[loops]
    else:
        strght_sizes = esize
        loop_sizes = [esize]*len(self_loops)

    if decimate_connections > 1:
        strght_edges = \
            strght_edges[::decimate_connections]

        if nonstring_container(esize):
            strght_sizes = \
                strght_sizes[::decimate_connections]

        if ecolor is not None:
            if nonstring_container(ecolor):
                strght_colors.extend(ecolor[strght][::decimate_connections])
            else:
                strght_colors.extend(
                    [ecolor] * (len(strght_edges) // decimate_connections))
    elif ecolor is not None:
        if nonstring_container(ecolor):
            strght_colors.extend(ecolor[strght])
        else:
            strght_colors.extend([ecolor]*len(strght_edges))

    return strght_edges, self_loops, strght_sizes, loop_sizes



#!/usr/bin/env python
#-*- coding:utf-8 -*-

# base_test.py

# This file is part of the NNGT module
# Distributed as a free software, in the hope that it will be useful, under the
# terms of the GNU General Public License.

import logging
import unittest
import xml.etree.ElementTree as xmlet

import nngt
from nngt.lib.test_functions import mpi_checker
from nngt.lib.logger import _log_message


logger = logging.getLogger()


# --------- #
# Xml tools #
# --------- #   

def _bool_from_string(string):
    return True if (string.lower() == "true") else False


def _list_from_string(string, elt_type, di_convert):
    lst = string[1:-1].split(", ")
    return [ di_convert[elt_type](elt) for elt in lst ]


def _xml_to_dict(xml_elt, di_types):
    di_result = {}
    for child in xml_elt:
        str_type = child.tag
        if len(child):
            name = str_type
            elt = child.find("start")
            if elt is not None:
                start = di_types[str_type](child.find("start").text)
                stop = di_types[str_type](child.find("stop").text)
                step = di_types[str_type](child.find("step").text)
                di_result[name] = np.arange(start,stop,step)
            else:
                di_result[child.tag] = _xml_to_dict(child, di_types)
        else:
            name = child.attrib["name"]
            if child.tag in ("string", "str"):
                text = child.text.replace("\\t","\t")
                di_result[name] = di_types[str_type](text)
            elif child.tag == "list":
                elt_type = child.attrib["type"]
                di_result[name] = _list_from_string( child.text,
                                                     elt_type, di_types )
            else:
                di_result[name] = di_types[str_type](child.text)
    return di_result


# ------------------------------------- #
# Decorator: repeat test for each graph #
# ------------------------------------- #

def foreach_graph(func):
    '''
    Decorator that automatically does the test for all graph instructions
    provided in the argument.
    '''
    def wrapper(*args, **kwargs):
        self = args[0]
        partial_backend = nngt.get_config("backend") in ("networkx", "nngt")
        for graph_name in self.graphs:
            if partial_backend and "corr" in graph_name:
                _log_message(logger, "DEBUG",
                             "Skipping correlated attributes with "
                             "networkx and nngt backends.")
            else:
                generated = self.gen_graph(graph_name)
                # check for None when using MPI
                if generated is not None:
                    g, di = generated
                    func(self, g, instructions=di, **kwargs)
    return wrapper
